import asyncio
import json
import os
import re
import threading
import time
from typing import List, Dict, Tuple, Optional, Any

import aiofiles
import numpy as np
from tqdm.asyncio import tqdm

from minion.configs.config import config
from minion.main.brain import Brain
from minion.main.rpyc_python_env import RpycPythonEnv
from minion.providers import create_llm_provider
from minion.providers.cost import CostManager
from minion.utils.syncheck import run_with_timeout


# Load JSONL file
def load_json(file_path):
    with open(file_path, "r") as f:
        data = json.load(f)
    return data


# Load JSONL file
def load_jsonl(file_path):
    data = []
    with open(file_path, "r") as f:
        for line in f:
            data.append(json.loads(line.strip()))
    return data


def extract_answer(answer_str):
    # Regular expression to find the answer after '####'
    match = re.search(r"####\s*(.*)", answer_str)
    if match:
        return match.group(1).strip()  # Extract and remove any surrounding whitespace
    else:
        return answer_str  # Return None if no match is found

async def evaluate_dataset(
    data,
    last_processed_id=None,
    start_id=None,
    to_processed_id=None,
    route="cot",
    run_filename=None,
    continue_process=False,
    concurrency_count=1,
):
    correct = 0
    count = 0
    total_count = len(data)
    matched_ids = []
    mismatch = []
    tasks = []

    async def process_batch(tasks, correct):
        results = await asyncio.gather(*tasks)
        for result in results:
            correct += result["result"]
            if result["result"] == 1:
                matched_ids.append(result["item_id"])
            else:
                mismatch.append(result)
        last_processed_item = results[-1]  # Get the last processed item
        return correct, last_processed_item

    async def read_json_file(filename):
        async with aiofiles.open(filename, "r") as f:
            contents = await f.read()
            data = json.loads(contents)
        return data

    async def save_run_info(filename, last_processed_id):
        run_info = {
            "last_processed_id": last_processed_id,
            "matched_ids": matched_ids,
            "mismatched_ids": mismatch,
            "correct": correct,
            "count": count,
            "correct_percentage": correct / count if count > 0 else 0,
            "total_prompt_tokens": cost_manager.total_prompt_tokens,
            "total_completion_tokens": cost_manager.total_completion_tokens,
            "total_cost": cost_manager.total_cost,
        }
        async with aiofiles.open(filename, "w") as f:
            await f.write(json.dumps(run_info, indent=4))

    if continue_process and os.path.exists(run_filename):
        async with aiofiles.open(run_filename, "r") as f:
            run_info = json.loads(await f.read())
        last_processed_id = run_info["last_processed_id"]
        matched_ids = run_info["matched_ids"]
        mismatch = run_info["mismatched_ids"]
        correct = run_info["correct"]
        count = run_info["count"]
        cost_manager.total_prompt_tokens = run_info.get("total_prompt_tokens", 0)
        cost_manager.total_completion_tokens = run_info.get("total_completion_tokens", 0)
        cost_manager.total_cost = run_info.get("total_cost", 0)

    with tqdm(total=total_count, desc="Evaluating") as pbar:
        for i, item in enumerate(data):
            item_id = i
            item["idx"] = i
            if last_processed_id is not None and item_id <= last_processed_id:
                continue
            if start_id and item_id < start_id:
                continue
            if to_processed_id and item_id > to_processed_id:
                break

            count += 1
            tasks.append(solve_single_question(item, route=route))

            if len(tasks) == concurrency_count:
                correct, last_processed_item = await process_batch(tasks, correct)
                last_processed_id = last_processed_item["item_id"]
                tasks = []  # Reset tasks after processing
                pbar.set_postfix({"Correct": correct, "count": count})
                pbar.update(concurrency_count)

                # Save running information after each batch
                await save_run_info(filename=run_filename, last_processed_id=last_processed_id)

        # Process remaining tasks
        if tasks:
            correct, last_processed_item = await process_batch(tasks, correct)
            last_processed_id = last_processed_item["item_id"]
            pbar.set_postfix({"Correct": correct})
            pbar.update(len(tasks))

            # Save running information after each batch
            await save_run_info(filename=run_filename, last_processed_id=last_processed_id)

    return correct, count, matched_ids, mismatch

PASS = "PASS"
FAIL = "FAIL"

def check_solution(solution, test, entry_point):
    print(f"solution: {solution}")

    try:
        # Define a global dictionary containing all necessary modules
        global_dict = {
            'math': __import__('math'),
            'hashlib': __import__('hashlib'),
            're': __import__('re'),
            'List': List,
            'Dict': Dict,
            'Tuple': Tuple,
            'Optional': Optional,
            'Any': Any
        }
        if entry_point == "decode_cyclic":
            solution = "\n\ndef encode_cyclic(s: str):\n    \"\"\"\n    returns encoded string by cycling groups of three characters.\n    \"\"\"\n    # split string to groups. Each of length 3.\n    groups = [s[(3 * i):min((3 * i + 3), len(s))] for i in range((len(s) + 2) // 3)]\n    # cycle elements in each group. Unless group has fewer elements than 3.\n    groups = [(group[1:] + group[0]) if len(group) == 3 else group for group in groups]\n    return \"\".join(groups)" + "\n\n" + solution
        elif entry_point == "decode_shift":
            solution = "\n\ndef encode_shift(s: str):\n    \"\"\"\n    returns encoded string by shifting every character by 5 in the alphabet.\n    \"\"\"\n    return \"\".join([chr(((ord(ch) + 5 - ord(\"a\")) % 26) + ord(\"a\")) for ch in s])\n\n\n" + solution
        elif entry_point == "find_zero":
            solution = "\n\ndef poly(xs: list, x: float):\n    return sum(coeff * (x ** i) for i, coeff in enumerate(xs))\n\n" + solution
        # Execute the solution
        exec(solution, global_dict)

        # Ensure the entry point function is defined
        if entry_point not in global_dict:
            raise ValueError(f"Function {entry_point} is not defined in the solution.")

        # Execute test cases
        exec(test, global_dict)

        # Get the check function
        check = global_dict["check"]

        # Run the check function with a timeout of 5 seconds
        result = run_with_timeout(check, (global_dict[entry_point],), 120)

        if result is None:
            result = (PASS, "Solution passed all test cases.")

    except TimeoutError:
        result = (FAIL, "Execution timeout. Please check if your solution contains infinite loops or time-consuming operations.")
    except Exception as e:
        # Record detailed error information
        error_message = f"Error: {str(e)}.\n Solution: {solution}.\n Test: {test}"
        result = (FAIL, error_message)

        # Write error information to error.log file
        with open('error.log', 'a', encoding='utf-8') as log_file:
            log_file.write(f"{time.strftime('%Y-%m-%d %H:%M:%S')} - {error_message}\n")

    return result

async def solve_single_question(item, route="nouse"):
    question = item["prompt"]
    canonical_solution = item["canonical_solution"]
    entry_point = item["entry_point"]
    test = item["test"]
    item_id = item.get("idx", -1)

    answer = await solve_question(item)

    ret = check_solution(answer, test, entry_point)
    if ret[0] == PASS:
        return {"result": 1, "item_id": item_id, "question": question, "answer": answer, "idx": item_id}

    else:
        # Append the mismatched item to the JSONL file
        return {
            "result": 0,
            "item_id": item_id,
            "question": question,
            "task_id": item["task_id"],
            "canonical_solution": canonical_solution,
            "test": test,
            "answer": answer,
            "reason": ret[1],
            "idx": item_id,
        }


async def solve_question(item):
    brain = Brain(stats_storer=None, python_env=RpycPythonEnv(ports=3007), llm=llm)
    current_dir = os.path.dirname(os.path.abspath(__file__))
    ensemble_logic_path = os.path.join(current_dir, "human_eval_config.json")
    # 加载测试用例
    test_cases_path = os.path.join(current_dir, "humaneval_public_test.jsonl")
    test_cases = load_jsonl(test_cases_path)
    # 查找对应的测试用例
    metadata = {"test_cases": []}
    for test_case in test_cases:
        if test_case["problem_id"] == item["task_id"]:
            metadata["test_cases"] = test_case.get("test", [])
            break
    answer, score, *_ = await brain.step(
        query="""Please provide a complete function implementation including:
- Full function definition
- All necessary logic
- Proper return statement
- Handle all edge cases

Here is the function to implement:
""" + item["prompt"],
    entry_point=item["entry_point"],
        dataset="HumanEval",
        execution_config=load_execution_config(ensemble_logic_path),
        metadata=metadata
    )
    return answer


# Load ensemble logic from JSON files
def load_execution_config(file_path):
    with open(file_path, "r") as file:
        ensemble_logic = json.load(file)
    return ensemble_logic

model = "default"
#model = "gpt-4o-mini"
#model = "gpt-4o"
llm = create_llm_provider(config.models.get(model))
cost_manager = CostManager()
llm.cost_manager = cost_manager
async def main():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    
    # 加载原始数据集
    file_name = os.path.join(current_dir, "human_eval_test.jsonl")
    original_data = load_jsonl(file_name)
    
    # 加载包含错误信息的 JSON 文件
    error_file = os.path.join(current_dir, "run_human_eval_deepseek0.json")
    with open(error_file, 'r') as f:
        error_data = json.load(f)
    
    # 从错误数据中提取 mismatched_ids
    mismatched_data = []
    for item in error_data["mismatched_ids"]:
        idx = item["idx"]
        # 从原始数据集中获取对应的完整数据
        if idx < len(original_data):
            mismatched_data.append(original_data[idx])

    # 使用新的数据集运行评估
    correct, count, matched_ids, mismatched_ids = await evaluate_dataset(
        mismatched_data, 
        run_filename=f"run_human_eval_ldb_{model}0.json",
        continue_process=True, 
        concurrency_count=60
    )

    print(f"Accuracy: {correct/count:.2%}")
    print(f"Mismatched IDs: {mismatched_ids}")


# Run the async main function
if __name__ == "__main__":
    asyncio.run(main())
# Example usage



from typing import Dict, Any, List, Optional, Tuple, Union, TYPE_CHECKING
from dataclasses import dataclass, field

if TYPE_CHECKING:
    from ..main.tool_hooks import HookConfig
import uuid
import asyncio
import logging
import inspect

from ..providers import BaseProvider
from ..tools.base_tool import BaseTool
from ..main.brain import Brain
from ..main.input import Input
from ..main.action_step import ActionStep, StreamChunk, StreamingActionManager
from minion.types.agent_response import AgentResponse
from minion.types.agent_state import AgentState
from minion.types.llm_types import ModelType, create_llm_from_model
from minion.utils.model_price import get_model_context_window, DEFAULT_CONTEXT_WINDOW
from minion.utils.token_counter import num_tokens_from_messages
from minion.types.history import History

logger = logging.getLogger(__name__)


@dataclass
class BaseAgent:
    """Agent基类，定义所有Agent的基本接口，支持生命周期管理"""
    
    name: str = "base_agent"
    tools: List[BaseTool] = field(default_factory=list)
    brain: Optional[Brain] = None
    hooks: Optional["HookConfig"] = None  # Pre/post tool use hooks for permission control

    # LLM configuration with strong typing support
    llm: Optional[Union[BaseProvider, str, ModelType]] = None  # Primary LLM provider, model name, or ModelType
    llms: Optional[Dict[str, Union[BaseProvider, str, ModelType]]] = None  # Multiple LLMs keyed by string
    
    system_prompt: Optional[str] = None  # 系统提示

    state: AgentState = field(default_factory=AgentState)
    user_id: Optional[str] = None
    agent_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    session_id: str = field(default_factory=lambda: str(uuid.uuid4()))
    max_steps: int = 20

    # Auto compact configuration
    auto_compact_enabled: bool = True           # Enable/disable auto compact
    auto_compact_threshold: float = 0.92        # Trigger threshold (92% of context window)
    auto_compact_keep_recent: int = 10          # Keep recent N messages as-is
    default_context_window: int = 128000        # Default context window size (128K tokens)
    compact_model: Optional[str] = None         # Model for compaction, None uses agent.llm

    # History decay configuration (for large outputs)
    decay_enabled: bool = True                  # Enable/disable history decay
    decay_ttl_steps: int = 3                    # Decay large outputs after N steps
    decay_min_size: int = 100_000               # Min size (bytes) to mark as decayable (100KB)
    decay_cache_dir: Optional[str] = None       # Cache dir, None uses ~/.minion/decay-cache

    # 生命周期管理
    _is_setup: bool = field(default=False, init=False)
    _toolsets: List[Any] = field(default_factory=list, init=False)  # List of toolset instances (MCP, UTCP, etc.)
    
    # 流式输出管理
    _streaming_manager: StreamingActionManager = field(default_factory=StreamingActionManager, init=False)
    
    def __post_init__(self):
        """初始化后的处理"""
        # Set agent reference in state
        if self.state and not self.state.agent:
            self.state.agent = self
            
        # Convert LLM configurations to BaseProvider instances
        self._convert_llm_configs()
            
        # Automatically handle toolset objects in tools parameter
        # not quite useful
        # current just for setting self._toolsets
        # and ensure await self._toolsets.ensure_setup() in setup()
        self._extract_toolsets_from_tools()

    def _convert_llm_configs(self):
        """Convert string/ModelType LLM configurations to BaseProvider instances"""
        # Convert primary LLM
        if self.llm is not None and not isinstance(self.llm, BaseProvider):
            self.llm = create_llm_from_model(self.llm)
            
        # Convert multiple LLMs
        if self.llms is not None:
            converted_llms = {}
            for key, model in self.llms.items():
                if isinstance(model, BaseProvider):
                    converted_llms[key] = model
                else:
                    converted_llms[key] = create_llm_from_model(model)
            self.llms = converted_llms

    @classmethod
    async def create(cls, *args, **kwargs) :
        """
        异步创建并设置实例
        
        Args:
            *args: 传递给构造函数的位置参数
            **kwargs: 传递给构造函数的关键字参数
            
        Returns:
            instance: 已设置完成的实例
        """
        instance = cls(*args, **kwargs)
        await instance.setup()
        return instance

    #what does this method do, not quite useful?
    def _extract_toolsets_from_tools(self):
        """
        Extract toolset objects from tools parameter and move them to _toolsets.
        This enables direct passing of toolset objects (MCP, UTCP, etc.) in the tools parameter.
        """
        if self._is_setup:
            raise RuntimeError("Cannot modify toolsets after setup")
            
        # Check if any tools are toolset objects (MCP, UTCP, etc.)
        toolsets = []
        regular_tools = []
        
        for tool in self.tools:
            # Check if it's a toolset object (has setup methods and connection/config params)
            if (hasattr(tool, 'ensure_setup') and hasattr(tool, 'connection_params')) or \
               (hasattr(tool, 'ensure_setup') and hasattr(tool, 'config')):
                toolsets.append(tool)
            else:
                regular_tools.append(tool)
        
        # Update the tools list to only contain regular tools
        self.tools = regular_tools #will add toolsets later if toolsets are healthy
        
        # Add toolset objects to _toolsets
        for toolset in toolsets:
            self._toolsets.append(toolset)
            logger.info(f"Auto-detected toolset {getattr(toolset, 'name', 'unnamed')} from tools parameter")
    
    async def setup(self):
        """Setup agent with tools"""
        # Prevent duplicate setup
        if self._is_setup:
            return
            
        # Setup toolsets (MCP, UTCP, etc.)
        for toolset in self._toolsets:
            try:
                await toolset.ensure_setup()
                if not toolset.is_healthy:
                    logger.warning(f"Toolset {toolset.name} failed to setup: {toolset.setup_error}")
            except Exception as e:
                logger.error(f"Failed to setup toolset {toolset.name}: {e}")

        # Get tools from healthy toolsets and merge into self.tools
        toolset_tools = []
        for toolset in self._toolsets:
            if toolset.is_healthy:
                toolset_tools.extend(toolset.get_tools())
            else:
                logger.warning(f"Skipping unhealthy toolset {toolset.name}")
        
        # Merge toolset tools into self.tools
        if toolset_tools:
            self.tools.extend(toolset_tools)
            logger.info(f"Added {len(toolset_tools)} toolset tools to agent")

        # Apply hooks to all tools if configured
        if self.hooks and (self.hooks.pre_tool_use or self.hooks.post_tool_use):
            from ..main.tool_hooks import wrap_tools_with_hooks
            logger.info(f"Applying hooks to {len(self.tools)} tools")
            self.tools = wrap_tools_with_hooks(self.tools, self.hooks)

        # Auto-convert raw functions to appropriate tool types
        self._convert_raw_functions_to_tools()
        
        # Wrap state-aware tools to automatically pass agent state
        logger.info("About to wrap state-aware tools...")
        self._wrap_state_aware_tools()
        logger.info("Finished wrapping state-aware tools")

        # Initialize brain with tools
        if self.brain is None:
            #we don't pass tools to brain, instead we
            #pass agent.tools when run/run_sync to brain.step
            #this way we can easily refresh tools on agent
            brain_kwargs = {'tools': [], 'state': self.state}
            
            # Handle LLMs (already converted to BaseProvider in __post_init__)
            if self.llm is not None:
                brain_kwargs['llm'] = self.llm
                
            if self.llms is not None:
                brain_kwargs['llms'] = self.llms
                    
            self.brain = Brain(**brain_kwargs)
        
        # Mark agent as setup
        self._is_setup = True
    
    async def close(self):
        """
        Agent清理关闭，在停止使用agent时调用
        这里会清理所有的toolset和其他资源
        """
        if not self._is_setup:
            logger.warning(f"Agent {self.name} not setup, skipping cleanup")
            return
        
        logger.info(f"Closing agent {self.name}")
        
        # 清理所有toolsets
        for toolset in self._toolsets:
            try:
                await toolset.close()
                logger.info(f"Closed toolset {getattr(toolset, 'name', 'unnamed')}")
            except Exception as e:
                logger.error(f"Error closing toolset {getattr(toolset, 'name', 'unnamed')}: {e}")
        
        # 清理toolset相关的工具
        self.tools = [tool for tool in self.tools if not self._is_toolset_tool(tool)]
        self._toolsets.clear()
        
        self._is_setup = False
        logger.info(f"Agent {self.name} cleanup completed")
    
    def _convert_raw_functions_to_tools(self):
        """
        自动将原始函数转换为相应的工具类型
        - 同步函数转换为BaseTool
        - 异步函数转换为AsyncBaseTool
        """
        from ..tools.tool_decorator import tool
        from ..tools.async_base_tool import AsyncBaseTool
        
        converted_tools = []
        conversion_count = 0
        
        for item in self.tools:
            # 检查是否是原始函数（不是工具实例）
            # 使用更通用的判断：如果是可调用对象但没有工具的基本属性，则认为是原始函数
            if callable(item) and not (hasattr(item, 'name') and hasattr(item, 'description')):
                try:
                    # 使用统一的tool装饰器进行转换
                    converted_tool = tool(item)
                    converted_tools.append(converted_tool)
                    conversion_count += 1
                    
                    # 记录转换信息
                    tool_type = "AsyncBaseTool" if isinstance(converted_tool, AsyncBaseTool) else "BaseTool"
                    logger.info(f"Auto-converted function '{item.__name__}' to {tool_type}")
                    
                except Exception as e:
                    logger.warning(f"Failed to convert function '{getattr(item, '__name__', str(item))}' to tool: {e}")
                    # 保留原始函数，不进行转换
                    converted_tools.append(item)
            else:
                # 保留已经是工具实例或具有工具属性的项目
                converted_tools.append(item)
        
        # 更新工具列表
        self.tools = converted_tools
        
        if conversion_count > 0:
            logger.info(f"Successfully auto-converted {conversion_count} raw functions to tools")
    
    def _wrap_state_aware_tools(self):
        """
        包装需要state的工具，使其能够接收agent的state
        """
        from ..tools.base_tool import BaseTool
        from ..tools.async_base_tool import AsyncBaseTool
        from ..tools.agent_state_aware_tool import AgentStateAwareTool
        import asyncio
        
        wrapped_tools = []
        wrap_count = 0
        
        for tool in self.tools:
            if hasattr(tool, 'needs_state') and tool.needs_state:

                tool_type = type(tool).__name__
                is_async = isinstance(tool, AsyncBaseTool)
                is_base = isinstance(tool, BaseTool)
                logger.debug(f"Wrapping tool {tool.name}: type={tool_type}, is_async={is_async}, is_base={is_base}")

                if isinstance(tool, AsyncBaseTool):
                    # 异步工具包装器 - 使用默认参数来捕获变量
                    def create_async_wrapper(original_forward, agent_ref):
                        async def wrapped_async_forward(self_tool, *args, **kwargs):
                            # 获取agent的state
                            agent_state = getattr(agent_ref, 'state', None)
                            # 将state作为关键字参数传递，这样不依赖于参数位置
                            return await original_forward(*args, state=agent_state, **kwargs)
                        return wrapped_async_forward

                    wrapper = create_async_wrapper(tool.forward, self)
                    tool.forward = wrapper.__get__(tool, type(tool))

                elif isinstance(tool, BaseTool):
                    # 同步工具包装器 - 使用默认参数来捕获变量
                    def create_sync_wrapper(original_forward, agent_ref):
                        def wrapped_sync_forward(self_tool, *args, **kwargs):
                            # 获取agent的state
                            agent_state = getattr(agent_ref, 'state', None)
                            # 将state作为第一个位置参数传递，不传递self_tool
                            return original_forward(*args, state=agent_state, **kwargs)
                        return wrapped_sync_forward

                    wrapper = create_sync_wrapper(tool.forward, self)
                    tool.forward = wrapper.__get__(tool, type(tool))

                wrap_count += 1
                logger.info(f"Successfully wrapped legacy state-aware tool: {tool.name}")
            
            wrapped_tools.append(tool)
        
        # 更新工具列表
        self.tools = wrapped_tools
        
        if wrap_count > 0:
            logger.info(f"Successfully wrapped {wrap_count} state-aware tools")

    def _is_toolset_tool(self, tool: BaseTool) -> bool:
        """检查工具是否是toolset工具"""
        return tool.__class__.__name__ in ['AsyncMcpTool', 'AsyncUtcpTool']
    
    async def __aenter__(self):
        """支持async context manager"""
        await self.setup()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """支持async context manager"""
        await self.close()
    

    
    @property
    def is_setup(self) -> bool:
        """检查agent是否已设置"""
        return self._is_setup
    
    def _ensure_setup(self):
        """确保agent已设置"""
        if not self._is_setup:
            raise RuntimeError(f"Agent {self.name} not setup. Call setup() first.")

    def run(self, 
           task: Optional[Union[str, Input]] = None,
           state: Optional[AgentState] = None, 
           max_steps: Optional[int] = None,
           reset: bool = False,
           llm: Optional[Union[str, BaseProvider]] = None,
           route: Optional[str] = None,
           **kwargs) -> Any:
        """
        Synchronous interface for running the agent.
        
        Args:
            task: Task description or Input object (required if state is None)
            state: Existing state for resuming interrupted execution
            max_steps: Maximum number of steps
            reset: If True, reset the agent state before execution
            llm: 可选的LLM名称（如 "code", "math", "creative"）或 BaseProvider 实例
            route: 可选的route名称，如 "code", "cot", "plan" 等，指定使用哪个minion
            **kwargs: Additional parameters
            
        Returns:
            Final task result
        """
        import asyncio
        return asyncio.run(self.run_async(task=task, state=state, max_steps=max_steps, reset=reset, stream=False, llm=llm, route=route, **kwargs))

    async def run_async(self, 
                       task: Optional[Union[str, Input]] = None,
                       state: Optional[AgentState] = None, 
                       max_steps: Optional[int] = None,
                       reset: bool = False,
                       stream: bool = False,
                       llm: Optional[Union[str, BaseProvider]] = None,
                       route: Optional[str] = None,
                       **kwargs) -> Any:
        """
        运行完整任务，自动多步执行直到完成，支持从中断状态恢复
        
        Args:
            task: 任务描述或Input对象 (state为None时必须提供)
            state: 已有状态，用于恢复中断的执行 (强类型AgentState)
            max_steps: 最大步数
            reset: If True, reset the agent state before execution
            stream: 若为True则使用异步迭代器返回中间结果
            llm: 可选的LLM名称（如 "code", "math", "creative"）或 BaseProvider 实例
            route: 可选的route名称，如 "code", "cot", "plan" 等，指定使用哪个minion
            **kwargs: 附加参数，可包含:
                - tools: 临时工具覆盖
                - 其他参数会传递给brain.step
                
        Returns:
            最终任务结果
        """
        self._ensure_setup()
        
        # 处理参数
        streaming = stream
        
        # 解析可选的LLM参数
        selected_llm_provider = None
        if llm is not None:
            selected_llm_provider = self._resolve_llm(llm)
            if selected_llm_provider is None:
                raise ValueError(f"LLM '{llm}' not found. Available LLMs: {list(self.list_available_llms().keys())}")
        
        # 处理状态初始化或恢复
        if state is None:
            if task is None:
                raise ValueError("Either 'task' or 'state' must be provided")
            # 初始化新状态
            self._init_state_from_task(task, route=route, **kwargs)
        else:
            # 使用已有状态
            self.state = state
                
            # 可选择更新task
            if task is not None:
                if isinstance(task, str):
                    self.state.task = task
                    # 更新input对象的query
                    if self.state.input and hasattr(self.state.input, 'query'):
                        self.state.input.query = task
                else:
                    self.state.task = task.query
                    self.state.input = task
            
            # 设置route（如果提供）
            if route is not None and self.state.input:
                self.state.input.route = route
        
        # 确定最大步数
        max_steps = max_steps or self.max_steps
        
        if stream:
            # 设置 stream_outputs 属性，供 UI 使用
            self.stream_outputs = True
            # 返回异步迭代器 - 使用 self.state 而不是原始 state 参数
            return self._run_stream(self.state, max_steps, kwargs, selected_llm_provider)
        else:
            # 清除 stream_outputs 属性
            self.stream_outputs = False
            # 一次性执行完成返回最终结果 - 使用 self.state 而不是原始 state 参数
            return await self._run_complete(self.state, max_steps, kwargs, selected_llm_provider)

    async def run_stream(self, 
                        task: Optional[Union[str, Input]] = None,
                        state: Optional[Dict[str, Any]] = None, 
                        max_steps: Optional[int] = None,
                        **kwargs):
        """
        Streaming interface for running the agent.
        
        Args:
            task: Task description or Input object (required if state is None)
            state: Existing state for resuming interrupted execution
            max_steps: Maximum number of steps
            **kwargs: Additional parameters
            
        Returns:
            AsyncGenerator: Stream of responses
        """
        # Force stream=True for this method
        return await self.run_async(task=task, state=state, max_steps=max_steps, stream=True, **kwargs)
            
    async def _run_stream(self, state, max_steps, kwargs, selected_llm_provider=None):
        """返回一个异步迭代器，逐步执行并返回中间结果"""
        step_count = 0
        
        while step_count < max_steps:
            # 开始新的步骤
            input_query = ""
            if state.input and hasattr(state.input, 'query'):
                input_query = str(state.input.query) if state.input.query else ""
            
            action_step = self._streaming_manager.start_step(
                step_type="reasoning",
                input_query=input_query
            )
            #
            # # yield 步骤开始信息
            # yield StreamChunk(
            #     content=f"[STEP {step_count + 1}] Starting reasoning...\n",
            #     chunk_type="step_start",
            #     metadata={"step_number": step_count + 1, "step_id": action_step.step_id}
            # )
            
            # 执行步骤并流式输出
            step_kwargs = kwargs.copy()
            if selected_llm_provider is not None:
                step_kwargs['llm'] = selected_llm_provider

            should_break = False
            async for chunk in self._execute_step_stream(state, **step_kwargs):
                action_step.add_chunk(chunk)
                yield chunk
                if hasattr(chunk,'is_final_answer') and chunk.is_final_answer:
                    action_step.is_final_answer = True
                    should_break = True  # Mark to break after streaming completes

            # 完成步骤
            result = action_step.to_agent_response()

            # If final_answer was called during streaming, terminate
            if should_break or action_step.is_final_answer:
                self._streaming_manager.complete_current_step(is_final_answer=True)
                break
            
            # 检查是否完成
            if self.is_done(result, state):
                action_step.is_final_answer = True
                self._streaming_manager.complete_current_step(is_final_answer=True)
                
                # yield StreamChunk(
                #     content=f"[FINAL] Task completed!\n",
                #     chunk_type="completion", #interesting, yield completion type
                #     metadata={"final_answer": True}
                # )
                break
            
            # 更新状态，继续下一步
            self._streaming_manager.complete_current_step()
            state = await self.update_state(state, result)
            step_count += 1
            
            # yield StreamChunk(
            #     content=f"\n[STEP {step_count}] Completed. Moving to next step...\n",
            #     chunk_type="step_end",
            #     metadata={"step_number": step_count}
            # )
            
        # 达到最大步数
        if step_count >= max_steps:
            # yield StreamChunk(
            #     content=f"\n[WARNING] Reached maximum steps ({max_steps}). Providing best available answer...\n",
            #     chunk_type="warning"
            # )
            
            try:
                final_answer = await self.provide_final_answer(state)
                yield StreamChunk(
                    content=f"Final answer: {final_answer}\n",
                    chunk_type="final_answer"
                )
            except Exception as e:
                logger.error(f"Failed to provide final answer: {e}")
                raise
    
    async def _execute_step_stream(self, state, **kwargs):
        """执行单个步骤的流式输出"""
        try:
            # 使用agent的tools
            tools = self.tools
            
            # 调用 brain.step 并检查是否返回流式生成器
            result = await self.brain.step(state, tools=tools, stream=True, system_prompt=self.system_prompt, **kwargs)
            
            # 如果 brain.step 返回的是异步生成器，则流式处理
            if inspect.isasyncgen(result):
                async for chunk in result:
                    if isinstance(chunk, str):
                        yield StreamChunk(content=chunk, chunk_type="llm_output")
                    elif isinstance(chunk, StreamChunk):
                        yield chunk
                    else:
                        yield StreamChunk(content=str(chunk), chunk_type="llm_output")
            else:
                # 如果不是流式，直接返回结果
                content = result.answer if hasattr(result, 'answer') else str(result)
                yield StreamChunk(content=content, chunk_type="llm_output")
                
        except Exception as e:
            logger.error(f"Error in step execution: {e}")
            raise
            
    async def _run_complete(self, state, max_steps, kwargs, selected_llm_provider=None):
        """一次性执行所有步骤直到完成，返回最终结果"""
        step_count = 0
        final_result = None
        
        while step_count < max_steps:
            # 传递selected_llm_provider给step方法
            step_kwargs = kwargs.copy()
            if selected_llm_provider is not None:
                step_kwargs['llm'] = selected_llm_provider
            result = await self.step(state, stream=kwargs.get('stream', False), **step_kwargs)
            
            # 检查是否完成
            if self.is_done(result, state):
                return result

            # 更新状态，继续下一步
            state = await self.update_state(state, result)
            step_count += 1

        # 达到最大步数
        if step_count >= max_steps:
            # Try to get the final answer and return it
            try:
                return await self.provide_final_answer(state)
            except Exception as e:
                # If getting the final answer fails, throw the original exception
                logger.error(f"Failed to provide final answer: {e}")
                raise Exception(f"Task execution reached max steps {max_steps} and is still incomplete")
            
        return final_result
    
    async def step(self, state: AgentState, stream: bool = False, llm: Optional[Union[str, BaseProvider]] = None, **kwargs) -> AgentResponse:
        """
        执行单步决策/行动
        Args:
            state: 强类型状态对象，包含 input 等必要信息
            llm: 可选的LLM名称（如 "code", "math", "creative"）或 BaseProvider 实例
            **kwargs: 其他参数，直接传递给brain
        Returns:
            AgentResponse: 结构化的响应对象
        """
        if not state.input:
            raise ValueError("State must contain input object")
            
        input_obj = state.input
        if not isinstance(input_obj, Input):
            # 将字符串转为Input
            input_obj = Input(query=str(input_obj))
            state.input = input_obj
            
        # 预处理输入
        input_obj, kwargs = await self.pre_step(input_obj, kwargs)
        
        # 解析可选的LLM参数
        if llm is not None:
            selected_llm_provider = self._resolve_llm(llm)
            if selected_llm_provider is None:
                raise ValueError(f"LLM '{llm}' not found.")
            kwargs['llm'] = selected_llm_provider
            
        # 执行主要步骤
        result = await self.execute_step(state, stream=stream, **kwargs)
        
        # 确保result是AgentResponse格式
        if not isinstance(result, AgentResponse):
            # 如果是旧的5-tuple格式，转换为AgentResponse
            result = AgentResponse.from_tuple(result)
        
        # 执行后处理操作
        await self.post_step(state.input, result)
        
        return result
    
    async def execute_step(self, state: AgentState, stream: bool = False, **kwargs) -> AgentResponse:
        """
        执行实际的步骤操作，默认委托给brain处理。子类可以重写此方法以自定义执行逻辑。
        Args:
            state: 强类型状态对象
            **kwargs: 其他参数
        Returns:
            AgentResponse: 结构化的响应对象
        """
        # 使用agent的tools
        tools = self.tools
        
        # 同步state到brain，这样minion可以访问agent的状态
        self.brain.state = state
        
        # 传递强类型状态给brain.step
        result = await self.brain.step(state, tools=tools, stream=stream, system_prompt=self.system_prompt, **kwargs)
        
        # 确保返回AgentResponse格式
        if not isinstance(result, AgentResponse):
            result = AgentResponse.from_tuple(result)
        
        return result
    
    async def pre_step(self, input_data: Input, kwargs: Dict[str, Any]) -> Tuple[Input, Dict[str, Any]]:
        """
        step执行前的预处理操作
        Args:
            input_data: 输入数据
            kwargs: 其他参数
        Returns:
            Tuple[Input, Dict[str, Any]]: 处理后的输入数据和参数
        """
        return input_data, kwargs
    
    async def post_step(self, input_data: Input, result: Any) -> None:
        """
        step执行后的处理操作
        Args:
            input_data: 输入数据
            result: step的执行结果 (可以是5-tuple或AgentResponse)
        """
        # 默认实现不执行任何操作
        pass
        
    async def provide_final_answer(self, state: AgentState) -> Any:
        """
        Attempt to provide a final answer when maximum steps are reached
        
        Args:
            state: Current agent state
            
        Returns:
            Final answer result
        """
        # 获取任务和历史
        task = state.task or ''
        history = state.history or []
        input_obj = state.input
        
        # 构建提示，要求LLM基于目前进展提供最终答案
        if isinstance(input_obj, Input):
            final_answer_prompt = f"""
You have reached the maximum step limit, but the task is not yet complete.

Original task: {task}

You have executed {len(history)} steps. Based on the current progress, please provide the best possible final answer or conclusion.
Even if the answer is not perfect, provide the best result you can currently derive.

Please provide the answer directly, without explaining why you couldn't complete the entire task.
"""
            # Update the input object's query
            input_obj.query = final_answer_prompt
            
            # Execute one step to get the final answer
            try:
                result = await self.step(state, stream=False)
                
                # Mark the result as final answer
                if hasattr(result, 'terminated') and not result.terminated:
                    # If using AgentResponse format
                    result.terminated = True
                    
                elif isinstance(result, tuple) and len(result) >= 3:
                    # If using 5-tuple format, modify the terminated flag
                    response, score, _, truncated, info = result
                    result = (response, score, True, truncated, info)
                    
                return result
            except Exception as e:
                logger.error(f"获取最终答案失败: {e}")
                # 构造一个基本的回应
                return f"The task could not be completed within the maximum step limit, but {len(history)} steps have been executed. Consider increasing the maximum step limit or simplifying the task."
        
        # If there is no valid input object, return basic information
        return f"The task execution reached the maximum step limit and could not provide a valid final answer."
    
    def _init_state_from_task(self, task: Union[str, Input], route: Optional[str] = None, **kwargs) -> None:
        """
        从任务初始化内部状态
        Args:
            task: 任务描述或Input对象
            route: 可选的route名称，指定使用哪个minion
            **kwargs: 附加参数
        """
        # 将任务转换为Input对象
        if isinstance(task, str):
            input_obj = Input(query=task)
            task_str = task
        else:
            input_obj = task
            task_str = task.query
        
        # 设置route（如果提供）
        if route is not None:
            input_obj.route = route
            
        # 初始化强类型状态
        from minion.types.history import History
        self.state = AgentState(
            agent=self,
            input=input_obj,
            history=History(),
            step_count=0,
            task=task_str
        )
        
        # 添加额外的metadata
        self.state.metadata.update(kwargs)

    async def update_state(self, state: AgentState, result: Any) -> AgentState:
        """
        Update state based on step result.

        Args:
            state: Current state (strongly typed AgentState)
            result: Step execution result (can be 5-tuple or AgentResponse)

        Returns:
            AgentState: Updated state
        """
        # Use the passed state
        self.state = state

        # Convert result to message format and add to history
        message_dict = self._convert_result_to_message(result)

        # Mark large content as decayable for future decay
        if self.decay_enabled:
            message_dict = self._maybe_mark_decayable(message_dict)

        self.state.history.append(message_dict)
        self.state.step_count += 1

        if isinstance(self.state.input, Input):
            self.state.input.query = f"Continue your task: {self.state.task}"

        # History decay check (decay old large outputs to files)
        if self.decay_enabled:
            self._check_and_decay()

        # Auto compact check
        if self.auto_compact_enabled and self._should_compact(self.state.history):
            self.state.history = await self._compact_history(self.state.history)

        return self.state
    
    def _convert_result_to_message(self, result: Any) -> Dict[str, Any]:
        """
        将步骤结果转换为OpenAI消息格式
        Args:
            result: 步骤执行结果 (可以是5-tuple或AgentResponse)
        Returns:
            Dict[str, Any]: OpenAI消息格式的字典
        """
        from minion.types.agent_response import AgentResponse
        
        # 如果是AgentResponse对象
        if isinstance(result, AgentResponse):
            content = result.content or str(result.raw_response) if result.raw_response is not None else ""
            message = {
                "role": "assistant",
                "content": content,
                "metadata": {
                    "score": result.score,
                    "confidence": result.confidence,
                    "terminated": result.terminated,
                    "truncated": result.truncated,
                    "is_final_answer": result.is_final_answer,
                    "timestamp": result.timestamp
                }
            }
            
            # 添加答案信息
            if result.answer is not None:
                message["metadata"]["answer"] = result.answer
            
            # 添加错误信息
            if result.error:
                message["metadata"]["error"] = result.error
                
            # 添加执行统计
            if result.execution_time:
                message["metadata"]["execution_time"] = result.execution_time
            if result.tokens_used:
                message["metadata"]["tokens_used"] = result.tokens_used
            message.pop("metadata")
            return message
        
        # 如果是5-tuple格式
        elif isinstance(result, tuple) and len(result) >= 5:
            response, score, terminated, truncated, info = result
            content = str(response) if response is not None else ""
            
            message = {
                "role": "assistant", 
                "content": content,
                "metadata": {
                    "score": score,
                    "terminated": terminated,
                    "truncated": truncated,
                    "info": info if isinstance(info, dict) else {}
                }
            }
            
            # 从info中提取特殊字段
            if isinstance(info, dict):
                if "answer" in info:
                    message["metadata"]["answer"] = info["answer"]
                if "is_final_answer" in info:
                    message["metadata"]["is_final_answer"] = info["is_final_answer"]
                if "confidence" in info:
                    message["metadata"]["confidence"] = info["confidence"]
                if "error" in info:
                    message["metadata"]["error"] = info["error"]
                    
            return message
        
        # 其他格式，直接转换为字符串内容
        else:
            return {
                "role": "assistant",
                "content": str(result),
                "metadata": {}
            }

    # ============================================
    # Auto Compact Methods
    # ============================================

    def _get_model_name(self) -> Optional[str]:
        """Get the model name from the LLM provider."""
        if self.llm is None:
            return None
        if hasattr(self.llm, 'config') and hasattr(self.llm.config, 'model'):
            return self.llm.config.model
        if hasattr(self.llm, 'model'):
            return self.llm.model
        return None

    def _get_context_window_limit(self) -> int:
        """Get current model's context window size.

        Returns:
            int: The context window limit in tokens.
        """
        model_name = self._get_model_name()
        if model_name:
            context_info = get_model_context_window(model_name)
            return context_info.get("max_input_tokens", self.default_context_window)
        return self.default_context_window

    def _calculate_current_tokens(self, history: History) -> int:
        """Calculate current history token count.

        Args:
            history: The conversation history.

        Returns:
            int: The total token count of the history.
        """
        if not history or len(history) == 0:
            return 0
        # Convert History to list of message dicts for token counting
        messages = list(history)
        model_name = self._get_model_name() or "gpt-4o"
        return num_tokens_from_messages(messages, model=model_name)

    def _should_compact(self, history: History) -> bool:
        """Check if compaction is needed based on threshold.

        Args:
            history: The conversation history.

        Returns:
            bool: True if compaction should be triggered.
        """
        if not self.auto_compact_enabled:
            return False
        if not history or len(history) <= self.auto_compact_keep_recent:
            return False

        current_tokens = self._calculate_current_tokens(history)
        context_limit = self._get_context_window_limit()
        threshold_tokens = int(context_limit * self.auto_compact_threshold)

        if current_tokens >= threshold_tokens:
            logger.info(
                f"Auto compact triggered: {current_tokens} tokens >= {threshold_tokens} "
                f"({self.auto_compact_threshold*100:.0f}% of {context_limit})"
            )
            return True
        return False

    def _get_compact_llm(self) -> BaseProvider:
        """Get LLM for compaction.

        Returns:
            BaseProvider: The LLM to use for generating summaries.
        """
        if self.compact_model:
            return create_llm_from_model(self.compact_model)
        return self.llm

    async def _compact_history(self, history: History) -> History:
        """Execute compaction - summarize old messages with LLM.

        Args:
            history: The conversation history to compact.

        Returns:
            History: The compacted history with summary + recent messages.
        """
        if len(history) <= self.auto_compact_keep_recent:
            return history

        # Split: old messages vs recent messages
        old_messages = list(history)[:-self.auto_compact_keep_recent]
        recent_messages = list(history)[-self.auto_compact_keep_recent:]

        # Separate system messages (always preserve)
        system_messages = [m for m in old_messages if m.get("role") == "system"]
        messages_to_summarize = [m for m in old_messages if m.get("role") != "system"]

        if not messages_to_summarize:
            return history

        # Build summary prompt
        summary_prompt = self._build_compact_prompt(messages_to_summarize)

        # Get LLM for compaction
        compact_llm = self._get_compact_llm()
        if compact_llm is None:
            logger.warning("No LLM available for compaction, skipping")
            return history

        try:
            # Call LLM to generate summary
            summary_messages = [
                {
                    "role": "system",
                    "content": (
                        "You are a conversation history summarizer. Compress the following "
                        "conversation history into a concise summary, preserving:\n"
                        "1. The user's main task and goals\n"
                        "2. Key steps that have been completed\n"
                        "3. Important code snippets or decisions made\n"
                        "4. Problems encountered and their solutions\n\n"
                        "The summary should be detailed enough to allow the conversation to "
                        "continue without losing context."
                    )
                },
                {"role": "user", "content": summary_prompt}
            ]

            # Use the LLM provider's generate method
            if hasattr(compact_llm, 'generate'):
                response = await compact_llm.generate(summary_messages)
                summary_content = response.get("content", "") if isinstance(response, dict) else str(response)
            elif hasattr(compact_llm, 'chat'):
                response = await compact_llm.chat(summary_messages)
                summary_content = response.get("content", "") if isinstance(response, dict) else str(response)
            else:
                logger.warning("LLM provider does not have generate or chat method, skipping compaction")
                return history

            # Build new history: system messages + summary + recent messages
            new_history = History()

            # Add preserved system messages
            for msg in system_messages:
                new_history.append(msg)

            # Add summary as a system message
            summary_message = {
                "role": "system",
                "content": (
                    f"[Conversation History Summary]\n{summary_content}\n"
                    "[End of Summary - Recent messages follow]"
                )
            }
            new_history.append(summary_message)

            # Add recent messages
            for msg in recent_messages:
                new_history.append(msg)

            logger.info(
                f"Compacted history from {len(history)} messages to {len(new_history)} messages "
                f"(summarized {len(messages_to_summarize)} messages)"
            )
            return new_history

        except Exception as e:
            logger.error(f"Failed to compact history: {e}")
            return history

    def _build_compact_prompt(self, messages: List[Dict[str, Any]]) -> str:
        """Build the prompt for summarizing messages.

        Args:
            messages: List of messages to summarize.

        Returns:
            str: The formatted prompt for summarization.
        """
        formatted_messages = []
        for msg in messages:
            role = msg.get("role", "unknown")
            content = msg.get("content", "")
            # Truncate very long messages in the summary request
            if len(content) > 2000:
                content = content[:2000] + "... [truncated]"
            formatted_messages.append(f"[{role.upper()}]: {content}")

        return (
            "Please summarize the following conversation history:\n\n"
            + "\n\n".join(formatted_messages)
            + "\n\nOutput the summary in a clear, structured format."
        )

    async def compact_now(self, state: Optional[AgentState] = None) -> None:
        """Manually trigger compaction.

        Args:
            state: The agent state to compact. Uses self.state if not provided.
        """
        target_state = state or self.state
        if target_state and hasattr(target_state, 'history'):
            target_state.history = await self._compact_history(target_state.history)
            logger.info("Manual compaction completed")

    # ============ History Decay Methods ============

    def _maybe_mark_decayable(self, message: Dict[str, Any]) -> Dict[str, Any]:
        """Mark large content as decayable for future decay.

        Args:
            message: The message to potentially mark

        Returns:
            Dict: The message, possibly with _decay_meta added
        """
        content = message.get("content", "")
        if not content or not isinstance(content, str):
            return message

        content_size = len(content.encode('utf-8'))
        if content_size >= self.decay_min_size:
            message["_decay_meta"] = {
                "step_created": self.state.step_count,
                "content_size": content_size,
                "decayable": True
            }
            logger.debug(f"Marked message as decayable: {content_size/1024:.0f}KB at step {self.state.step_count}")

        return message

    def _get_decay_cache_dir(self) -> "Path":
        """Get the cache directory for decayed outputs."""
        from pathlib import Path
        if self.decay_cache_dir:
            cache_dir = Path(self.decay_cache_dir)
        else:
            cache_dir = Path.home() / '.minion' / 'decay-cache'
        cache_dir.mkdir(parents=True, exist_ok=True)
        return cache_dir

    def _save_decay_content(self, content: str, step_created: int) -> str:
        """Save content to file and return the file path.

        Args:
            content: The content to save
            step_created: The step number when content was created

        Returns:
            str: Absolute path to the saved file
        """
        cache_dir = self._get_decay_cache_dir()
        file_id = str(uuid.uuid4())[:8]
        filename = f"decay-step{step_created}-{file_id}.txt"
        file_path = cache_dir / filename
        file_path.write_text(content, encoding='utf-8')
        return str(file_path)

    def _decay_message(self, msg: Dict[str, Any]) -> Dict[str, Any]:
        """Decay a message by saving content to file and replacing with reference.

        Args:
            msg: The message to decay (must have _decay_meta)

        Returns:
            Dict: New message with file reference instead of full content
        """
        content = msg.get("content", "")
        meta = msg.get("_decay_meta", {})

        # Save content to file
        file_path = self._save_decay_content(content, meta.get("step_created", 0))

        # Build replacement content
        size_kb = meta.get("content_size", len(content.encode('utf-8'))) / 1024
        new_content = (
            f"[Large output ({size_kb:.0f}KB) saved to: {file_path}]\n"
            f"Use file_read to access full content if needed."
        )

        logger.info(f"Decayed message from step {meta.get('step_created')}: {size_kb:.0f}KB -> {file_path}")

        return {
            "role": msg.get("role", "user"),
            "content": new_content,
            "_decayed": True,
            "_decay_file": file_path,
            "_decay_original_size": meta.get("content_size", 0)
        }

    def _check_and_decay(self) -> int:
        """Check and decay expired large outputs in history.

        Returns:
            int: Number of messages decayed
        """
        if not self.decay_enabled:
            return 0

        if not hasattr(self.state, 'history') or not self.state.history:
            return 0

        current_step = self.state.step_count
        decayed_count = 0

        # Iterate through history and decay expired large outputs
        for i, msg in enumerate(self.state.history):
            meta = msg.get("_decay_meta")
            if not meta or not meta.get("decayable"):
                continue

            # Already decayed
            if msg.get("_decayed"):
                continue

            step_age = current_step - meta.get("step_created", 0)
            if step_age >= self.decay_ttl_steps:
                self.state.history[i] = self._decay_message(msg)
                decayed_count += 1

        if decayed_count > 0:
            logger.info(f"Decayed {decayed_count} large outputs at step {current_step}")

        return decayed_count

    def is_done(self, result: Any, state: AgentState) -> bool:
        """
        判断任务是否完成
        Args:
            result: 当前步骤结果 (可以是5-tuple或AgentResponse)
            state: 当前状态 (强类型AgentState)
        Returns:
            bool: 是否完成
        """
        # 检查AgentResponse类型
        if hasattr(result, 'is_done') and callable(result.is_done):
            return result.is_done()
        elif hasattr(result, 'terminated') or hasattr(result, 'is_final_answer'):
            return getattr(result, 'terminated', False) or getattr(result, 'is_final_answer', False)

        # 检查5-tuple格式
        if isinstance(result, tuple) and len(result) >= 3:
            # 返回格式为 (response, score, terminated, truncated, info)
            terminated = result[2]
            return terminated

        # 检查状态中的final_answer标志
        if state.is_final_answer:
            return True

        # 检查结果中是否包含 FINAL_ANSWER 标记（来自 final_answer 工具调用）
        if hasattr(result, 'answer') and result.answer:
            answer = str(result.answer)
            if 'FINAL_ANSWER:' in answer:
                return True

        return False

    def _get_builtin_meta_tools(self) -> Dict[str, Any]:
        """获取内置的meta工具"""
        from ..tools.think_tool import ThinkTool
        from ..tools.meta_tools import PlanTool, ReflectionTool
        
        return {
            "think": ThinkTool(),
            "plan": PlanTool(),
            "reflect": ReflectionTool(),
        }
    
    def add_meta_tools(self) -> None:
        """添加内置的meta工具到agent"""
        meta_tools = self._get_builtin_meta_tools()
        for tool_name, tool in meta_tools.items():
            self.add_tool(tool)
            logger.info(f"Added meta tool: {tool_name}")
    
    def add_tool(self, tool: BaseTool) -> None:
        """
        添加工具
        Args:
            tool: 要添加的工具
        """
        self.tools.append(tool)
        # 同时向brain添加工具
        if hasattr(self.brain, 'add_tool'):
            self.brain.add_tool(tool)
    
    def set_primary_llm(self, model: Union[ModelType, str, BaseProvider]) -> None:
        """
        Set the primary LLM for the agent
        
        Args:
            model: Model type, string identifier, or provider instance
        """
        # Convert to BaseProvider if needed
        if isinstance(model, BaseProvider):
            self.llm = model
        else:
            self.llm = create_llm_from_model(model)
            
        # Update brain if already setup
        if self._is_setup and self.brain:
            self.brain.llm = self.llm
    
    def add_specialized_llm(self, name: str, model: Union[ModelType, str, BaseProvider]) -> None:
        """
        Add a specialized LLM for specific tasks
        
        Args:
            name: Name/key for the specialized LLM
            model: Model type, string identifier, or provider instance
        """
        if self.llms is None:
            self.llms = {}
            
        # Convert to BaseProvider if needed
        if isinstance(model, BaseProvider):
            self.llms[name] = model
        else:
            self.llms[name] = create_llm_from_model(model)
        
        # Update brain if already setup
        if self._is_setup and self.brain:
            if not hasattr(self.brain, 'llms') or self.brain.llms is None:
                self.brain.llms = {}
            self.brain.llms[name] = self.llms[name]
    
    def _resolve_llm(self, llm: Union[str, BaseProvider]) -> Optional[BaseProvider]:
        """
        Resolve LLM name or provider to BaseProvider instance
        
        Args:
            llm: LLM name ("primary", "code", "math", etc.) or BaseProvider instance
            
        Returns:
            BaseProvider instance or None if not found
        """
        # If already a BaseProvider instance, return it directly
        if isinstance(llm, BaseProvider):
            return llm
        
        # Otherwise treat as string name
        if llm == "primary":
            return self.llm
        elif self.llms and llm in self.llms:
            return self.llms[llm]
        else:
            return None
    
    def get_tool(self, tool_name: str) -> Optional[BaseTool]:
        """
        获取指定名称的工具
        Args:
            tool_name: 工具名称
        Returns:
            找到的工具实例，如果未找到则返回None
        """
        for tool in self.tools:
            if tool.name == tool_name:
                return tool
        return None
    
    def add_memory(self, messages: Union[str, List[Dict[str, str]]], metadata: Optional[Dict[str, Any]] = None) -> None:
        """
        添加记忆到brain的mem0
        Args:
            messages: 要添加的消息内容，可以是字符串或消息列表
                     如果是字符串，会被转换为 [{"role": "user", "content": messages}]
                     如果是列表，每个消息都应该是 {"role": xxx, "content": xxx} 格式
            metadata: 额外的元数据
        """
        if self.brain and self.brain.mem:
            # 确保至少有一个必需的ID
            if not any([self.user_id, self.agent_id, self.session_id]):
                raise ValueError("At least one of user_id, agent_id, or session_id (run_id) is required!")
                
            self.brain.mem.add(
                messages=messages,  # mem0会自动处理字符串到消息列表的转换
                user_id=self.user_id,
                agent_id=self.agent_id,
                run_id=self.session_id,
                metadata=metadata
            )
            
    def get_all_memories(self, filter: Optional[Dict[str, Any]] = None) -> List[Any]:
        """
        根据元数据检索记忆
        
        Args:
            filter: 可选的元数据过滤条件，用于筛选记忆

        Returns:
            List[Memory]: 检索到的记忆列表
        """
        if self.brain and self.brain.mem:
            response = self.brain.mem.get_all(
                user_id=self.user_id,
                agent_id=self.agent_id,
                run_id=self.session_id
            )
            
            # 处理返回的数据格式，提取results列表
            memories = []
            if isinstance(response, dict) and 'results' in response:
                # 新的API格式 (v1.1)
                memories = response['results']
            elif isinstance(response, list):
                # 旧的API格式
                memories = response
            else:
                return []
            
            # 如果提供了过滤条件，进行过滤
            if filter:
                filtered_memories = []
                for memory in memories:
                    if "metadata" in memory:
                        # 检查元数据是否满足过滤条件
                        match = True
                        for key, value in filter.items():
                            if key not in memory["metadata"] or memory["metadata"][key] != value:
                                match = False
                                break
                        if match:
                            filtered_memories.append(memory)
                return filtered_memories
            
            return memories
        return []
        
    def search_memories(self, query: str, top_k: int = 5, include_relations: bool = False) -> Union[List[Any], Dict[str, Any]]:
        """
        语义搜索记忆
        Args:
            query: 搜索查询
            top_k: 返回结果数量
            include_relations: 是否包含关系数据（仅在mem0 API v1.1中有效）
        Returns:
            如果include_relations=False: List[Memory]: 搜索结果列表
            如果include_relations=True: Dict[str, Any]: 包含'results'和'relations'的字典
        """
        if self.brain and self.brain.mem:
            try:
                # 尝试使用top_k参数
                response = self.brain.mem.search(
                    user_id=self.user_id,
                    agent_id=self.agent_id,
                    run_id=self.session_id,
                    query=query,
                    top_k=top_k
                )
            except TypeError:
                # 如果top_k参数不被接受，尝试使用limit参数
                try:
                    response = self.brain.mem.search(
                        user_id=self.user_id,
                        agent_id=self.agent_id,
                        run_id=self.session_id,
                        query=query,
                        limit=top_k
                    )
                except TypeError:
                    # 如果limit参数也不被接受，只使用必需参数
                    response = self.brain.mem.search(
                        user_id=self.user_id,
                        agent_id=self.agent_id,
                        run_id=self.session_id,
                        query=query
                    )
            
            # 处理返回的数据格式
            if isinstance(response, dict):
                if include_relations and 'relations' in response:
                    # 如果需要关系数据，直接返回完整响应
                    return response
                elif 'results' in response:
                    # 否则只返回结果列表
                    return response['results']
            elif isinstance(response, list):
                # 旧的API格式
                return response
                
        return []
        
    def get_conversation_history(self, top_k: int = 10, order: str = "desc") -> List[Dict[str, str]]:
        """
        获取记忆中的对话历史
        Args:
            top_k: 返回的消息数量
            order: 排序方式，"asc" 或 "desc"
        Returns:
            List[Dict[str, str]]: 对话历史列表
        """
        if self.brain and self.brain.mem:
            response = self.brain.mem.get_conversation_history(
                user_id=self.user_id,
                agent_id=self.agent_id,
                run_id=self.session_id,
                top_k=top_k,
                order=order
            )
            
            # 处理返回的数据格式
            if isinstance(response, dict) and 'results' in response:
                # 新的API格式 (v1.1)
                return response['results']
            elif isinstance(response, list):
                # 旧的API格式
                return response
                
        return []
    
    def get_state(self) -> Dict[str, Any]:
        """
        获取当前执行状态
        Returns:
            Dict[str, Any]: 当前状态的副本
        """
        if hasattr(self, 'state') and self.state:
            return self.state.copy()
        else:
            raise ValueError("No current state available. Run agent first.")
    
    def save_state(self, filepath: str) -> None:
        """
        保存状态到文件
        Args:
            filepath: 保存路径
        """
        import pickle
        state = self.get_state()
        with open(filepath, 'wb') as f:
            pickle.dump(state, f)
    
    def load_state(self, filepath: str) -> Dict[str, Any]:
        """
        从文件加载状态
        Args:
            filepath: 文件路径
        Returns:
            Dict[str, Any]: 加载的状态
        """
        import pickle
        with open(filepath, 'rb') as f:
            return pickle.load(f)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Skills system for Minion

This module provides a skills system inspired by Claude Skills,
allowing users to define, load, and execute specialized AI skills.
"""

from minion.skills.skill import Skill
from minion.skills.skill_registry import (
    SkillRegistry,
    get_skill_registry,
    reset_skill_registry,
)
from minion.skills.skill_loader import (
    SkillLoader,
    load_skills,
    get_available_skills,
)

__all__ = [
    "Skill",
    "SkillRegistry",
    "get_skill_registry",
    "reset_skill_registry",
    "SkillLoader",
    "load_skills",
    "get_available_skills",
]



#!/usr/bin/env python3
"""
Test script to verify that state-aware tools work correctly.
"""

import asyncio
import inspect
import sys
import os

import pytest

# Add the project root to the path
sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))

from minion.agents.code_agent import CodeAgent
from minion.types.agent_state import CodeAgentState
from minion.tools.tool_decorator import tool

# Create test tools with and without state parameter

@tool
def regular_tool(message: str) -> str:
    """A regular tool that doesn't need state.
    
    Args:
        message: The message to process
        
    Returns:
        Processed message
    """
    return f"Regular tool processed: {message}"

@tool
def state_aware_tool(action: str, state: CodeAgentState) -> str:
    """A tool that needs access to agent state.
    
    Args:
        state: The agent state
        action: The action to perform
        
    Returns:
        Action result with state info
    """
    step_count = state.step_count if state else 0
    history_len = len(state.history) if state and state.history else 0
    return f"State-aware tool: {action} (step={step_count}, history={history_len})"

@tool
async def async_state_aware_tool(task: str, state: CodeAgentState) -> str:
    """An async tool that needs access to agent state.
    
    Args:
        state: The agent state
        task: The task to perform
        
    Returns:
        Task result with state info
    """
    await asyncio.sleep(0.1)  # Simulate async work
    task_name = state.task if state else "unknown"
    return f"Async state-aware tool: {task} (current_task={task_name})"
@pytest.mark.asyncio
async def test_state_aware_tools():
    """Test that state-aware tools work correctly."""
    
    print("=== Testing State-Aware Tools ===\n")
    
    # Test 1: Check tool needs_state detection
    print("Test 1: Tool needs_state detection")
    
    print(f"regular_tool.needs_state: {regular_tool.needs_state}")
    print(f"state_aware_tool.needs_state: {state_aware_tool.needs_state}")
    print(f"async_state_aware_tool.needs_state: {async_state_aware_tool.needs_state}")
    
    assert regular_tool.needs_state == False, "Regular tool should not need state"
    assert state_aware_tool.needs_state == True, "State-aware tool should need state"
    assert async_state_aware_tool.needs_state == True, "Async state-aware tool should need state"
    
    print("✓ Tool needs_state detection works correctly")
    print()
    
    # Test 2: Test tools with agent
    print("Test 2: Tools with agent")
    
    # Create agent with our test tools
    agent = CodeAgent(
        name="test_agent",
        tools=[regular_tool, state_aware_tool, async_state_aware_tool],
        enable_state_tracking=False,
        use_async_executor=False
    )
    
    await agent.setup()
    
    # Set some state
    agent.state.step_count = 5
    agent.state.task = "Test task"
    agent.history.append("Step 1")
    agent.history.append("Step 2")
    
    print(f"Agent state: step_count={agent.state.step_count}, task={agent.state.task}, history_len={len(agent.history)}")
    
    # Test regular tool (should work without state)
    try:
        result = regular_tool("Hello world")
        print(f"Regular tool result: {result}")
        assert "Regular tool processed: Hello world" in result, "Regular tool should work normally"
        print("✓ Regular tool works correctly")
    except Exception as e:
        print(f"✗ Regular tool failed: {e}")
    
    print()
    
    # Test state-aware tool (should receive state automatically)
    try:
        # Find the wrapped tool in agent.tools
        wrapped_state_tool = None
        for tool in agent.tools:
            if tool.name == "state_aware_tool":
                wrapped_state_tool = tool
                break
        
        if wrapped_state_tool:
            result = wrapped_state_tool(action="test action")
            print(f"State-aware tool result: {result}")
            assert "step=5" in result, "State-aware tool should receive step_count"
            assert "history=2" in result, "State-aware tool should receive history length"
            print("✓ State-aware tool works correctly")
        else:
            print("✗ Could not find wrapped state-aware tool")
    except Exception as e:
        print(f"✗ State-aware tool failed: {e}")
    
    print()
    
    # Test async state-aware tool (should receive state automatically)
    try:
        # Find the wrapped async tool in agent.tools
        wrapped_async_tool = None
        for tool in agent.tools:
            if tool.name == "async_state_aware_tool":
                wrapped_async_tool = tool
                break
        
        if wrapped_async_tool:
            result = await wrapped_async_tool(task="async task")
            print(f"Async state-aware tool result: {result}")
            assert "current_task=Test task" in result, "Async state-aware tool should receive task"
            print("✓ Async state-aware tool works correctly")
        else:
            print("✗ Could not find wrapped async state-aware tool")
    except Exception as e:
        print(f"✗ Async state-aware tool failed: {e}")
    
    print()
    
    # Test 3: Test tool wrapping in agent setup
    print("Test 3: Tool wrapping in agent setup")
    
    # Check that tools were wrapped correctly
    wrapped_count = 0
    for tool in agent.tools:
        if hasattr(tool, 'needs_state') and tool.needs_state:
            wrapped_count += 1
            print(f"Found state-aware tool: {tool.name}")
    
    assert wrapped_count == 2, f"Should have 2 state-aware tools, found {wrapped_count}"
    print("✓ State-aware tools were wrapped correctly")
    
    print()
    
    # Test 4: Test state changes
    print("Test 4: State changes")
    print("Available tools:")
    for tool in agent.tools:
        print(f"  - {tool.name}, async={inspect.iscoroutinefunction(tool.forward)}")
    
    # Change agent state
    agent.state.step_count = 10
    agent.state.task = "Updated task"
    agent.history.append("Step 3")
    
    # Test that tools see the updated state
    try:
        wrapped_state_tool = None
        for tool in agent.tools:
            if tool.name == "state_aware_tool":
                wrapped_state_tool = tool
                break
        
        if wrapped_state_tool:
            result = wrapped_state_tool(action="updated action")
            print(f"Updated state result: {result}")
            assert "step=10" in result, "Tool should see updated step_count"
            assert "history=3" in result, "Tool should see updated history length"
            print("✓ Tools see updated state correctly")
        else:
            print("✗ Could not find wrapped tool for updated state test")
    except Exception as e:
        print(f"✗ Updated state test failed: {e}")
    
    print()
    
    # Test 5: Test without state (edge case)
    print("Test 5: Test without state (edge case)")
    
    # Temporarily remove state
    original_state = agent.state
    agent.state = None
    
    try:
        wrapped_state_tool = None
        for tool in agent.tools:
            if tool.name == "state_aware_tool":
                wrapped_state_tool = tool
                break
        
        if wrapped_state_tool:
            result = wrapped_state_tool(action="no state action")
            print(f"No state result: {result}")
            assert "step=0" in result, "Tool should handle missing state gracefully"
            print("✓ Tools handle missing state gracefully")
        else:
            print("✗ Could not find wrapped tool for no state test")
    except Exception as e:
        print(f"✗ No state test failed: {e}")
    finally:
        # Restore state
        agent.state = original_state
    
    print()
    
    print("=== State-Aware Tools Tests Completed! 🎉 ===")
    
    # Cleanup
    await agent.close()

if __name__ == "__main__":
    asyncio.run(test_state_aware_tools())


import logging
import asyncio
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Tuple

from minion.tools import AsyncBaseTool, Toolset

logger = logging.getLogger(__name__)

DEBUG = False  # Set to True for debug output


def format_tools_for_bedrock(tools: List[AsyncBaseTool]) -> Tuple[List[Dict[str, Any]], Dict[str, str]]:
    """
    Convert minion AsyncBaseTool tools to Bedrock tool format.
    
    Args:
        tools: List of minion AsyncBaseTool tools
        
    Returns:
        Tuple containing:
        - List of tools formatted for Bedrock
        - Mapping between modified tool names and original names
    """
    bedrock_tools = []
    tool_name_mapping = {}
    
    for tool in tools:
        # Create the input schema JSON from tool.inputs
        input_schema_json = {
            "type": "object",
            "properties": {},
            "required": []
        }
        
        # Add inputs to the input schema
        if hasattr(tool, 'inputs') and tool.inputs:
            input_schema_json["properties"] = tool.inputs
            # Extract required fields from tool inputs
            # In minion framework, we need to determine required fields
            # This could be based on tool signature or explicit marking
            required_fields = []
            for field_name, field_info in tool.inputs.items():
                # Check if field is marked as required (you may need to adjust this logic)
                if isinstance(field_info, dict) and not field_info.get("nullable", False):
                    required_fields.append(field_name)
            input_schema_json["required"] = required_fields
        
        # Replace periods in tool name with underscores
        original_name = tool.name
        bedrock_tool_name = original_name.replace(".", "_")
        
        # Truncate if longer than 64 characters (Bedrock's limit)
        if len(bedrock_tool_name) > 64:
            short_uuid = str(uuid.uuid4())[:8]
            short_name = f"{bedrock_tool_name[:55]}_{short_uuid}"
            if DEBUG:
                print(f"Tool name '{bedrock_tool_name}' is too long, using '{short_name}' instead")
            bedrock_tool_name = short_name
        
        # Store the mapping between the modified name and original name
        tool_name_mapping[bedrock_tool_name] = original_name
        
        # Format the tool for Bedrock
        tool_spec = {
            "name": bedrock_tool_name,
            "description": tool.description,
            "inputSchema": {
                "json": input_schema_json
            }
        }
        
        bedrock_tools.append({"toolSpec": tool_spec})
    
    return bedrock_tools, tool_name_mapping


class AsyncUtcpTool(AsyncBaseTool):
    """
    Adapter class to convert UTCP tools to minion framework compatible format
    """
    
    def __init__(self, name: str, description: str, utcp_tool, utcp_client, original_name: str):
        super().__init__()
        self.name = name.replace(".", "_")  # Replace dots with underscores for valid method names
        self.description = description
        self.utcp_client = utcp_client
        self.original_name = original_name  # Keep original name for UTCP calls
        self.utcp_tool = utcp_tool  # Store the original UTCP tool
        
        # Convert UTCP tool inputs to minion format
        self.inputs = self._convert_utcp_inputs_to_minion_format(utcp_tool)
        
        # Add attributes expected by minion framework
        self.__name__ = self.name
        self.__doc__ = description
        self.__input_schema__ = self.inputs
    
    def _convert_utcp_inputs_to_minion_format(self, utcp_tool) -> Dict[str, Dict[str, Any]]:
        """
        Convert UTCP tool inputs to minion framework format
        
        Args:
            utcp_tool: The original UTCP tool object
            
        Returns:
            Dict in minion format for tool inputs
        """
        minion_inputs = {}
        
        # Extract inputs from UTCP tool
        if hasattr(utcp_tool, 'inputs') and utcp_tool.inputs:
            if hasattr(utcp_tool.inputs, 'properties') and utcp_tool.inputs.properties:
                for field_name, field_info in utcp_tool.inputs.properties.items():
                    # Convert UTCP field format to minion format
                    minion_field = {}
                    
                    if hasattr(field_info, 'type'):
                        minion_field["type"] = field_info.type
                    if hasattr(field_info, 'description'):
                        minion_field["description"] = field_info.description
                    
                    # Check if field is required
                    required_fields = getattr(utcp_tool.inputs, 'required', [])
                    if field_name not in required_fields:
                        minion_field["nullable"] = True
                    
                    minion_inputs[field_name] = minion_field
        
        return minion_inputs

    async def forward(self, **kwargs) -> Dict[str, Any]:
        """Execute the UTCP tool with given parameters"""
        try:
            logger.info(f"Executing tool call: {self.original_name}")
            result = await self.utcp_client.call_tool(self.original_name, kwargs)
            logger.info(f"Tool execution successful!")
            logger.info(f"Result: {result}")
            
            # Format the tool result as expected by the framework
            return result
        except Exception as e:
            logger.error(f"Error executing UTCP tool {self.original_name}: {e}")
            return {"error": f"Error: {str(e)}"}

class UtcpManualToolset(Toolset):
    """
    UTCP toolset that follows Google ADK pattern.
    Can be passed directly in the tools parameter when creating an agent.
    """
    
    def __init__(
        self, 
        config: Optional[Union[str, Path, Dict[str, Any]]] = None,
        root_dir: Optional[str] = None,
        name: Optional[str] = None,
        setup_timeout: float = 30,  # 30 seconds timeout for setup
        ignore_setup_errors: bool = False,
    ):
        """
        Initialize UtcpManualToolset with configuration
        
        Args:
            config: UTCP configuration - can be a path to config file, dict config, or UtcpClientConfig
            root_dir: Root directory for UTCP client
            name: Optional name for the toolset
            setup_timeout: Timeout in seconds for toolset setup
            ignore_setup_errors: If True, setup errors will be logged but not raised
        """
        # Initialize with empty tools list first
        super().__init__([])
        
        self.config = config
        self.root_dir = root_dir
        self.name = name or f"utcp_toolset_{id(self)}"
        self._utcp_client = None
        self._is_setup = False
        self._setup_timeout = setup_timeout
        self._ignore_setup_errors = ignore_setup_errors
        self._setup_error: Optional[Exception] = None

    @property
    def is_healthy(self) -> bool:
        """Return True if the toolset is setup and healthy"""
        return self._is_setup and not self._setup_error

    @property
    def setup_error(self) -> Optional[Exception]:
        """Return the setup error if any"""
        return self._setup_error

    async def initialize_utcp_client(self):
        """Initialize the UTCP client with configuration."""
        try:
            from utcp.utcp_client import UtcpClient
            #from utcp.data.utcp_client_config import UtcpClientConfigSerializer
            from utcp.data.tool import Tool
        except ImportError as e:
            logger.error(f"UTCP library not available: {e}")
            raise RuntimeError(f"UTCP library not available: {e}")
        
        # Use provided config or default to providers.json in same directory
        config = self.config

        logger.info(f"Initializing UTCP client with config: {config}")
        client = await UtcpClient.create(
            root_dir=self.root_dir,
            config=config
        )
        return client

    async def setup(self):
        """Internal setup method that does the actual work"""
        # Initialize UTCP client
        self._utcp_client = await self.initialize_utcp_client()
        
        # Get available tools from UTCP client
        try:
            # Assuming UTCP client has a method to list available tools
            # This might need to be adjusted based on actual UTCP API
            tools_info = await self._utcp_client.search_tools("Please give me all tools")
            logger.info(f"Connected to UTCP with {len(tools_info)} tools")
            
            # Convert to AsyncUtcpTool objects
            self.tools = []
            for tool_info in tools_info:
                # Extract tool information - adjust based on actual UTCP tool format
                original_name = tool_info.name
                safe_name = original_name.replace(".", "_")  # Make name valid for method calls
                description = tool_info.description
                
                utcp_tool = AsyncUtcpTool(
                    name=safe_name,
                    description=description,
                    utcp_tool=tool_info,  # Pass the entire UTCP tool object
                    utcp_client=self._utcp_client,
                    original_name=original_name
                )
                self.tools.append(utcp_tool)
            
            logger.info(f"UtcpManualToolset '{self.name}' setup completed with {len(self.tools)} tools")
            
        except AttributeError:
            # If list_tools method doesn't exist, we'll need to handle this differently
            logger.warning("UTCP client doesn't have list_tools method, creating empty toolset")
            self.tools = []
        self._is_setup = True

    async def ensure_setup(self) -> None:
        """Ensure toolset is setup, with timeout handling"""
        if self._is_setup:
            return

        try:
            async with asyncio.timeout(self._setup_timeout):
                await self.setup()
                self._is_setup = True
                self._setup_error = None
        except Exception as e:
            self._setup_error = e
            logger.error(f"Failed to setup UtcpManualToolset {self.name}: {e}")
            if not self._ignore_setup_errors:
                raise

    def get_tools(self) -> List[AsyncBaseTool]:
        """
        Get list of tools, returns empty list if setup failed and errors are ignored
        """
        if not self.is_healthy:
            logger.warning(f"UtcpManualToolset {self.name} is not healthy, returning empty tool list")
            return []
        return self.tools

    async def close(self):
        """Close the toolset and clean up resources"""
        if self._utcp_client:
            # Close UTCP client if it has a close method
            if hasattr(self._utcp_client, 'close'):
                await self._utcp_client.close()
            self._utcp_client = None
            self._is_setup = False
        self.tools.clear()
        logger.info(f"UtcpManualToolset '{self.name}' closed")

    def __repr__(self):
        return f"UtcpManualToolset(name={self.name}, setup={self._is_setup})"


# Factory function for creating UTCP toolset
async def create_utcp_toolset(
    config: Optional[Union[str, Path, Dict[str, Any]]] = None,
    root_dir: Optional[str] = None,
    name: Optional[str] = None
) -> UtcpManualToolset:
    """
    Create a UTCP toolset and set it up
    
    Args:
        config: UTCP configuration - can be a path to config file, dict config, or UtcpClientConfig
        root_dir: Root directory for UTCP client
        name: Optional name for the toolset
        
    Returns:
        UtcpManualToolset configured and setup with the provided parameters
    """
    toolset = UtcpManualToolset(
        config=config,
        root_dir=root_dir,
        name=name or "utcp_toolset"
    )
    
    # Automatically setup the toolset
    await toolset.ensure_setup()
    
    return toolset


"""
Minion Types Module

This module contains type definitions for the Minion framework.
"""

from .agent_response import (
    AgentResponse,
    Usage,
    StreamChunk,
    UserMessage,
    AssistantMessage,
    ThinkingMessage,
    ToolUseMessage,
    ToolResultMessage,
    CodeExecutionMessage,
    SystemMessage,
    ResultMessage,
    Message,
)
from .agent_state import AgentState, CodeAgentState

__all__ = [
    'AgentResponse',
    'AgentState',
    'CodeAgentState',
    'Usage',
    'StreamChunk',
    'UserMessage',
    'AssistantMessage',
    'ThinkingMessage',
    'ToolUseMessage',
    'ToolResultMessage',
    'CodeExecutionMessage',
    'SystemMessage',
    'ResultMessage',
    'Message',
]



from minion.main.minion import Minion


class PreprocessingMinion(Minion):
    """
    PreprocessingMinion handles the transformation of raw input into refined perceptions.

    While named 'Preprocessing' for general understanding, this minion actually performs
    complex cognitive enhancements analogous to the perception process in the Active
    Inference Framework (AIF). It transforms raw observations into meaningful internal
    representations through various cognitive processes.
    """

    async def execute(self):
        if not self.input.execution_config or "preprocessing" not in self.input.execution_config:
            return self.input

        observations = self.collect_raw_observations()
        enhanced_perceptions = await self.enhance_cognitive_representations(observations)
        self.update_input_with_enhanced_perceptions(enhanced_perceptions)

        return self.input

    def collect_raw_observations(self):
        return {
            field: getattr(self.input, field)
            for field in ["query", "short_context", "instruction", "solution", "answer", "feedback"]
            if hasattr(self.input, field)
        }

    async def enhance_cognitive_representations(self, observations):
        enhanced_perceptions = observations.copy()
        for step in self.input.execution_config["preprocessing"]:
            if step["type"] == "re2":
                enhanced_perceptions = self.apply_attention_enhancement(enhanced_perceptions, step)
            elif step["type"] == "rephrase":
                enhanced_perceptions = await self.apply_semantic_refinement(enhanced_perceptions, step)

        return enhanced_perceptions

    def update_input_with_enhanced_perceptions(self, enhanced_perceptions):
        for field, value in enhanced_perceptions.items():
            setattr(self.input, field, value)

    def apply_attention_enhancement(self, perceptions, step):
        for field in step["apply_to"]:
            if field in perceptions:
                for _ in range(step.get("repeat", 1)):
                    perceptions[field] = self.apply_re2(field, perceptions[field])
        return perceptions

    async def apply_semantic_refinement(self, perceptions, step):
        for field in step["apply_to"]:
            if field in perceptions:
                perceptions[field] = await self.apply_rephrase(perceptions[field])
        return perceptions

    def apply_re2(self, field, text):
        if not text:
            return text
        return f"{text}\nRead the above again: {text}"

    async def apply_rephrase(self, text):
        if not text:
            return text
        rephrase_prompt = f"Rephrase the following without changing its meaning: {text}"
        rephrased_text = await self.brain.llm.aask(rephrase_prompt)
        return rephrased_text



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Multi Tool Use Module - 处理GPT生成的并行工具调用

这个模块拦截GPT生成的 `from multi_tool_use import parallel` 调用，
并提供实际的并行工具执行功能。
"""

import asyncio
import inspect
import concurrent.futures
import threading
from typing import List, Dict, Any, Union, Optional, Callable, Awaitable, TypeVar, Sequence


# 定义类型变量
T = TypeVar('T')  # 工具返回值类型
ToolUse = Dict[str, Any]  # 单个工具调用配置
Config = Union[Dict[str, List[ToolUse]], List[ToolUse], None]  # 配置格式类型
ToolResult = Dict[str, Any]  # 工具结果类型
Result = Dict[str, Union[List[ToolResult], int, bool]]  # 返回结果类型


async def parallel(config: Config) -> Result:
    """
    并行执行多个工具调用
    
    Args:
        config: 包含 tool_uses 列表的配置字典或直接的工具调用列表
        
    Returns:
        包含所有工具执行结果的字典
    """
    # Handle both dict and list formats
    if isinstance(config, list):
        tool_uses = config
    elif isinstance(config, dict):
        tool_uses = config.get("tool_uses", [])
    else:
        return {"results": [], "error": "Invalid config format: must be dict or list"}
        
    if not tool_uses:
        return {"results": [], "error": "No tool_uses provided"}
    
    # 获取当前执行环境中的工具
    frame = inspect.currentframe()
    tools_dict: Dict[str, Callable] = {}
    
    # 遍历调用栈，查找可用的工具
    while frame:
        local_vars = frame.f_locals
        global_vars = frame.f_globals
        
        # 优先查找 static_tools（AsyncPythonExecutor 中的工具）
        if 'static_tools' in local_vars:
            static_tools = local_vars['static_tools']
            if isinstance(static_tools, dict):
                # 直接查找 functions 命名空间
                if 'functions' in static_tools:
                    functions_ns = static_tools['functions']
                    if hasattr(functions_ns, '__dict__'):
                        for attr_name in dir(functions_ns):
                            if not attr_name.startswith('_'):
                                tool = getattr(functions_ns, attr_name)
                                if callable(tool):
                                    tools_dict[f"functions.{attr_name}"] = tool
                
                # 也查找其他工具
                for name, tool in static_tools.items():
                    if name == 'functions':
                        continue  # 已经处理过了
                    if callable(tool) and hasattr(tool, '__name__'):
                        tools_dict[f"functions.{tool.__name__}"] = tool
                    elif hasattr(tool, 'name') and hasattr(tool, 'forward'):
                        tools_dict[f"functions.{tool.name}"] = tool
                    elif callable(tool):
                        tools_dict[f"functions.{name}"] = tool
                        
        # 检查全局和局部变量中的 functions 命名空间
        for vars_dict in [local_vars, global_vars]:
            if 'functions' in vars_dict:
                functions_obj = vars_dict['functions']
                if hasattr(functions_obj, '__dict__') or hasattr(functions_obj, '__dir__'):
                    for attr_name in dir(functions_obj):
                        if not attr_name.startswith('_'):
                            attr_value = getattr(functions_obj, attr_name)
                            if callable(attr_value):
                                tools_dict[f"functions.{attr_name}"] = attr_value
        
        # 检查 __main__ 执行环境中的工具
        if '__name__' in local_vars and local_vars.get('__name__') == '__main__':
            for name, obj in local_vars.items():
                if name.startswith('async_') and callable(obj):
                    tools_dict[f"functions.{name}"] = obj
        
        # 查找工具函数和对象
        for name, obj in {**local_vars, **global_vars}.items():
            if callable(obj) and hasattr(obj, '__name__'):
                if obj.__name__.startswith('async_') or obj.__name__.endswith('_tool'):
                    tools_dict[f"functions.{obj.__name__}"] = obj
            elif hasattr(obj, 'name') and hasattr(obj, 'forward'):
                # BaseTool或AsyncBaseTool实例
                tools_dict[f"functions.{obj.name}"] = obj
        
        frame = frame.f_back
    
    # 准备并行执行的任务
    tasks: List[Awaitable[Any]] = []
    
    for tool_use in tool_uses:
        recipient_name = tool_use.get("recipient_name", "")
        parameters = tool_use.get("parameters", {})
        
        # 查找对应的工具
        tool = tools_dict.get(recipient_name)
        if not tool:
            # 尝试不带 functions. 前缀
            tool_name = recipient_name.replace("functions.", "")
            tool = tools_dict.get(f"functions.{tool_name}")
        
        if tool:
            # 创建工具调用任务
            task = _execute_single_tool(tool, parameters, recipient_name)
            tasks.append(task)
        else:
            # 工具未找到，添加错误结果
            tasks.append(_create_error_result(recipient_name, "Tool not found"))
    
    # 并行执行所有任务
    results = await asyncio.gather(*tasks, return_exceptions=True)
    
    # 格式化结果
    formatted_results: List[Dict[str, Any]] = []
    for i, result in enumerate(results):
        tool_use = tool_uses[i]
        if isinstance(result, Exception):
            formatted_results.append({
                "recipient_name": tool_use.get("recipient_name", ""),
                "error": str(result),
                "success": False
            })
        elif isinstance(result, dict) and "error" in result:
            # 处理错误结果（如 Tool not found）
            formatted_results.append({
                "recipient_name": tool_use.get("recipient_name", ""),
                "error": result.get("error", "Unknown error"),
                "success": False
            })
        else:
            formatted_results.append({
                "recipient_name": tool_use.get("recipient_name", ""),
                "result": result,
                "success": True
            })
    
    return {
        "results": formatted_results,
        "total_calls": len(tool_uses),
        "successful_calls": sum(1 for r in formatted_results if r.get("success", False)),
        "failed_calls": sum(1 for r in formatted_results if not r.get("success", False))
    }


async def _execute_single_tool(tool: Any, parameters: Dict[str, Any], tool_name: str) -> Any:
    """
    执行单个工具
    
    Args:
        tool: 工具对象或函数
        parameters: 参数字典
        tool_name: 工具名称
        
    Returns:
        工具执行结果
    """
    try:
        # 处理字符串类型的数值参数
        processed_params: Dict[str, Any] = {}
        for key, value in parameters.items():
            if isinstance(value, str):
                # 尝试转换数字字符串，但更保守一些
                if value.replace('.', '').replace('-', '').isdigit():
                    # 包含小数点的浮点数
                    if '.' in value:
                        try:
                            processed_params[key] = float(value)
                        except ValueError:
                            processed_params[key] = value
                    # 纯整数
                    else:
                        try:
                            processed_params[key] = int(value)
                        except ValueError:
                            processed_params[key] = value
                else:
                    # 保持为字符串
                    processed_params[key] = value
            else:
                processed_params[key] = value
        
        # 检查工具类型并调用
        if hasattr(tool, 'forward'):
            # BaseTool或AsyncBaseTool实例 - 应该调用工具实例本身，不是直接调用forward
            # 这样可以确保通过__call__方法，包含初始化检查和参数处理逻辑
            if asyncio.iscoroutinefunction(tool.forward):
                result = await tool(**processed_params)  # 调用__call__方法
            else:
                result = tool(**processed_params)  # 调用__call__方法
        elif callable(tool):
            # 普通函数或异步函数
            if asyncio.iscoroutinefunction(tool):
                result = await tool(**processed_params)
            else:
                result = tool(**processed_params)
        else:
            raise ValueError(f"Tool {tool_name} is not callable")
        
        # 添加调试信息，同时确保结果正确返回
        print(f"Tool {tool_name.replace('functions.', '')} execution result: {result}")
        return result
        
    except Exception as e:
        raise Exception(f"Error executing {tool_name}: {str(e)}")


async def _create_error_result(tool_name: str, error_message: str) -> Dict[str, Any]:
    """创建错误结果"""
    return {
        "error": error_message,
        "tool_name": tool_name
    }


def smart_parallel(config: Config = None, **kwargs: Any) -> Result:
    """
    智能的并行工具执行，自动检测异步环境和参数格式
    
    Args:
        config: 包含 tool_uses 列表的配置字典，或者直接传递工具列表
        **kwargs: 支持 tool_uses 等关键字参数
        
    Returns:
        包含所有工具执行结果的字典
    """
    # 处理不同的调用方式
    if config is None and 'tool_uses' in kwargs:
        # parallel(tool_uses=[...]) 调用方式
        config = {"tool_uses": kwargs['tool_uses']}
    elif isinstance(config, list):
        # parallel([...]) 调用方式，保持列表格式
        config = config
    elif not isinstance(config, (dict, list)):
        raise ValueError("参数必须是包含 'tool_uses' 键的字典，或者直接传递工具列表，或者使用 tool_uses 关键字参数")
    
    # 尝试同步执行，避免复杂的异步检测
    try:
        # 检查是否已经在事件循环中
        loop = asyncio.get_running_loop()
        # 在异步环境中，创建新的事件循环来执行
        
        def run_in_thread() -> Result:
            return asyncio.run(parallel(config))
        
        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(run_in_thread)
            return future.result()
            
    except RuntimeError:
        # 没有运行中的循环，直接同步执行
        return asyncio.run(parallel(config))


# 为了兼容性，也提供同步版本
def parallel_sync(config: Config) -> Result:
    """
    同步版本的并行工具执行（实际上会在内部使用异步）
    
    Args:
        config: 包含 tool_uses 列表的配置字典或直接的工具调用列表
        
    Returns:
        包含所有工具执行结果的字典
    """
    try:
        # 尝试获取当前事件循环
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # 如果已经在异步环境中，创建任务
            task = asyncio.create_task(parallel(config))
            return task
        else:
            # 如果不在异步环境中，运行异步函数
            return loop.run_until_complete(parallel(config))
    except RuntimeError:
        # 没有事件循环，创建新的
        return asyncio.run(parallel(config))


# 默认导出智能版本和异步版本
__all__ = ["parallel", "smart_parallel", "parallel_sync"]


from enum import Enum, auto
from typing import Dict, Type

class ImproveRoute(Enum):
    """改进路由的枚举类"""
    FEEDBACK = "feedback"
    REASONING = "reasoning"
    STEP_BY_STEP = "step_by_step"
    
    @classmethod
    def get_route(cls, route_name: str) -> "ImproveRoute":
        """根据字符串获取对应的改进路由
        
        Args:
            route_name: 路由名称字符串
            
        Returns:
            ImproveRoute: 匹配的改进路由，如果没有匹配项则返回默认的 FEEDBACK 路由
        """
        from minion.main.minion import IMPROVER_MINIONS
        route_name = route_name.lower()
        return IMPROVER_MINIONS.get(route_name, IMPROVER_MINIONS.get(cls.FEEDBACK))


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Chain of Thought worker minions
"""
from minion.actions.lmp_action_node import LmpActionNode
from minion.main.base_workers import WorkerMinion
from minion.main.minion import register_worker_minion
from minion.types.agent_response import AgentResponse
from minion.main.prompt import (
    COT_PROBLEM_INSTRUCTION,
    DCOT_PROMPT,
    WORKER_PROMPT,
    TASK_INPUT,
)
from minion.main.prompt import extract_think_and_answer
from minion.utils.answer_extraction import extract_answer
from minion.utils.template import construct_messages_from_template


@register_worker_minion
class CotMinion(WorkerMinion):
    """Chain of Thought (CoT) Strategy, Ask the LLM to think step-by-step, explaining each part of the problem to enhance the accuracy of the answer. Please noted you can't access web or user's local computer, so if you need information from the web or from user's local computer, DON'T USE THIS STRATEGY."""

    def __init__(self, worker_config=None, **kwargs):
        super().__init__(worker_config=worker_config, **kwargs)
        self.worker_config = worker_config
        self.input.instruction = "let's think step by step to solve this problem"

    async def execute(self):
        if self.task:
            # Task mode: use TASK_INPUT template
            template_str = COT_PROBLEM_INSTRUCTION + WORKER_PROMPT + TASK_INPUT
            messages = construct_messages_from_template(
                template_str, self.input, task=self.task
            )
        else:
            # Normal mode: use original prompt
            messages = construct_messages_from_template(
                COT_PROBLEM_INSTRUCTION + WORKER_PROMPT, self.input
            )

        # 优先使用selected_llm，如果没有则使用brain.llm
        llm_to_use = self.selected_llm if self.selected_llm else self.brain.llm
        node = LmpActionNode(llm_to_use)
        tools = (self.input.tools or []) + (self.brain.tools or [])
        response = await node.execute(messages, tools=tools)
        self.answer_node = node

        # Check post_processing setting, giving precedence to worker_config
        post_processing = None
        if self.worker_config and 'post_processing' in self.worker_config:
            post_processing = self.worker_config['post_processing']
        elif self.input.post_processing:
            post_processing = self.input.post_processing

        if post_processing == "extract_python" or self.input.query_type == "code_solution":
            self.answer = response  # Let route minion handle extraction
        else:
            # For DeepSeek think mode, extract the answer part outside <think> tags
            think_content, answer_content = extract_think_and_answer(response)
            self.answer = answer_content if answer_content else response
            # Store think content for potential debugging/analysis
            self.think_content = think_content

        self.input.answer = self.answer
        self.answer_raw = self.input.answer_raw = response

        # Return AgentResponse instead of just the answer
        return AgentResponse(
            raw_response=response,
            answer=self.answer,
            is_final_answer=True,
            score=1.0,
            terminated=True,
            truncated=False,
            info={'raw_response': response}
        )

    async def execute_stream(self):
        """流式执行方法"""
        if self.task:
            # Task mode: use TASK_INPUT template
            template_str = COT_PROBLEM_INSTRUCTION + WORKER_PROMPT + TASK_INPUT
            messages = construct_messages_from_template(
                template_str, self.input, task=self.task
            )
        else:
            # Normal mode: use original prompt
            messages = construct_messages_from_template(
                COT_PROBLEM_INSTRUCTION + WORKER_PROMPT, self.input
            )

        node = LmpActionNode(self.get_llm())
        tools = (self.input.tools or []) + (self.brain.tools or [])

        full_response = ""
        async for chunk in self.stream_node_execution(node, messages, tools):
            # Yield StreamChunk对象，保持原始结构
            yield chunk

            # 累积内容用于后续处理
            if hasattr(chunk, 'content'):
                full_response += chunk.content
            elif isinstance(chunk, str):
                full_response += chunk

        # 处理完整响应
        post_processing = None
        if self.worker_config and 'post_processing' in self.worker_config:
            post_processing = self.worker_config['post_processing']
        elif self.input.post_processing:
            post_processing = self.input.post_processing

        if post_processing == "extract_python" or self.input.query_type == "code_solution":
            self.answer = full_response  # Let route minion handle extraction
        else:
            # For DeepSeek think mode, extract the answer part outside <think> tags
            think_content, answer_content = extract_think_and_answer(full_response)
            self.answer = answer_content if answer_content else full_response
            # Store think content for potential debugging/analysis
            self.think_content = think_content

        self.input.answer = self.answer
        self.answer_raw = self.input.answer_raw = full_response


# https://x.com/_philschmid/status/1842846050320544016
class DcotMinion(WorkerMinion):
    """Dynamic Chain of Thought Strategy"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.input.instruction = ""

    async def execute(self):
        if self.task:
            # Task mode: use TASK_INPUT template
            template_str = DCOT_PROMPT + TASK_INPUT
            messages = construct_messages_from_template(
                template_str, self.input, task=self.task
            )
        else:
            # Normal mode: use original prompt
            messages = construct_messages_from_template(
                DCOT_PROMPT, self.input
            )

        node = LmpActionNode(self.get_llm())
        #tools = (self.input.tools or []) + (self.brain.tools or [])
        response = await node.execute(messages, tools=None)

        self.answer_node = node
        self.answer = self.input.answer = extract_answer(response)
        self.answer_raw = self.input.answer_raw = response

        # Return AgentResponse instead of just the answer
        return AgentResponse(
            response=self.answer,
            score=1.0,
            terminated=False,
            truncated=False,
            info={'raw_response': response}
        )

    async def execute_stream(self):
        """流式执行方法"""
        if self.task:
            # Task mode: use TASK_INPUT template
            template_str = DCOT_PROMPT + TASK_INPUT
            messages = construct_messages_from_template(
                template_str, self.input, task=self.task
            )
        else:
            # Normal mode: use original prompt
            messages = construct_messages_from_template(
                DCOT_PROMPT, self.input
            )

        node = LmpActionNode(self.get_llm())

        full_response = ""
        async for chunk in self.stream_node_execution(node, messages, tools=None):
            # Yield StreamChunk对象，保持原始结构
            yield chunk

            # 累积内容用于后续处理
            if hasattr(chunk, 'content'):
                full_response += chunk.content
            elif isinstance(chunk, str):
                full_response += chunk

        # 处理完整响应
        self.answer = self.input.answer = extract_answer(full_response)
        self.answer_raw = self.input.answer_raw = full_response



from typing import Any, Union, List, Optional, Type
from pydantic import BaseModel
import json
import xml.etree.ElementTree as ET
import re
import random
import inspect

from tenacity import retry, stop_after_attempt, retry_if_exception_type

from minion.configs.config import config
from minion.schema.message_types import Message
from minion.actions.action_node import LLMActionNode
from minion.schema.messages import user
from minion.providers import create_llm_provider
from minion.models.schemas import Answer  # Import the Answer model
from minion.exceptions import FinalAnswerException
from minion.tools.tool_decorator import tool as decorate_tool


def _format_tool_result(tool_name: str, tool_result: Any) -> str:
    """
    Format tool execution result for LLM consumption.

    Args:
        tool_name: Name of the tool
        tool_result: Result from tool execution

    Returns:
        Formatted string for the tool result
    """
    # Special handling for final_answer tool
    if tool_name == 'final_answer':
        if isinstance(tool_result, dict):
            result_str = json.dumps(tool_result, ensure_ascii=False, indent=2)
        else:
            result_str = str(tool_result)
        return f"FINAL_ANSWER:{result_str}"

    # Format the result based on type
    if isinstance(tool_result, dict):
        result_str = json.dumps(tool_result, ensure_ascii=False, indent=2)
    elif isinstance(tool_result, (list, tuple)):
        result_str = json.dumps(tool_result, ensure_ascii=False, indent=2)
    else:
        result_str = str(tool_result)

    return f"Tool {tool_name} execution result:\n{result_str}"


# @ell.complex(model="gpt-4o-mini")
# def ell_call(ret):
#     """You are a helpful assistant."""
#     return ret
class LmpActionNode(LLMActionNode):
    def __init__(self, llm, input_parser=None, output_parser=None):
        super().__init__(llm, input_parser, output_parser)
        #ell.init(**config.ell, default_client=self.llm.client_sync)

    async def execute(self, messages: Union[str, Message, List[Message], dict, List[dict]], response_format: Optional[Union[Type[BaseModel], dict]] = None, output_raw_parser=None, format="json", tools=None, tool_choice="auto", stream=False, stop=None, **kwargs) -> Any:
        # 处理 system_prompt 参数
        system_prompt = kwargs.pop('system_prompt', None)
        
        # 处理可选的 llm 参数
        selected_llm = kwargs.pop('llm', None)
        current_llm = selected_llm if selected_llm is not None else self.llm
        
        # 添加 input_parser 处理
        if self.input_parser:
            messages = self.input_parser(messages)

        # Convert string/single message to list
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]
        elif isinstance(messages, Message):
            # Convert Message object to dictionary format
            messages = [messages.model_dump()]
        elif isinstance(messages, dict) and "role" in messages:
            # Single message in dictionary format
            messages = [messages]
        elif isinstance(messages, list) and all(isinstance(msg, Message) for msg in messages):
            # Convert list of Message objects to list of dictionaries
            messages = [msg.model_dump() for msg in messages]

        # 如果有 system_prompt，将其添加到 messages 的开头
        if system_prompt:

            # has_system = any(msg.get("role") == "system" for msg in messages)
            # if not has_system:
                messages.insert(0, {"role": "system", "content": system_prompt})

        # 处理 tools 参数
        if tools is not None:
            # 将 tools 转换为 API 格式
            tools_formatted = self._format_tools_for_api(tools)
            if tools_formatted:
                api_params = {
                    "temperature": current_llm.config.temperature,
                    "model": current_llm.config.model,
                    "tools": tools_formatted,
                    "tool_choice": tool_choice,
                    "original_tools": tools  # 保存原始工具对象
                }
            else:
                api_params = {
                    "temperature": current_llm.config.temperature,
                    "model": current_llm.config.model,
                    "original_tools": tools  # 保存原始工具对象
                }
        else:
            # 从 llm.config 获取配置
            api_params = {
                "temperature": current_llm.config.temperature, #+ random.random() * 0.01, #add random to avoid prompt caching
                "model": current_llm.config.model,
            }

        # 处理stop参数
        if stop is not None:
            api_params['stop'] = stop
        
        # 将 kwargs 合并到 api_params 中，允许覆盖默认值
        api_params.update(kwargs)
        
        # 处理流式参数
        if stream:
            api_params['stream'] = True
            
        original_response_format = response_format

        if isinstance(response_format, type) and issubclass(response_format, BaseModel):
            # 生成 schema
            schema = response_format.model_json_schema()
            schema_with_indent = json.dumps(schema, indent=4)

            # 创建示例数据
            example = response_format.model_construct()

            if format == "json":
                example_str = example.model_dump_json(indent=4)
                prompt = (
                    f"Please provide the response in JSON format as per the following schema:\n"
                    f"{schema_with_indent}\n\n"
                    f"Here's an example of the expected format:\n"
                    f"{example_str}\n\n"
                    f"Please ensure your response follows this exact schema format."
                )
                api_params['response_format'] = { "type": "json_object" }
            else:  # format == "xml" or format == "xml_simple"
                example_dict = example.model_dump()
                example_xml = self._dict_to_xml_example(example_dict)
                prompt = (
                    f"""Construct an XML response that adheres to the specified schema below.

Schema Structure Example:
{example_xml}

Required JSON Schema Compliance:
{schema_with_indent}

Your response should be:

Well-formed XML: Ensure it follows XML syntax rules.
Schema-compliant: Each element, attribute, and data type must match the JSON schema requirements.
Error-free for Parsing: Escape all special characters and ensure compatibility for JSON conversion.
Provide a final XML structure that aligns seamlessly with both the XML and JSON schema constraints."""
                )
                api_params['response_format'] = { "type": "text" }

            messages.append({"role": "user", "content": prompt})

        # 根据是否流式调用不同的方法
        if stream:
            # 流式模式：返回异步生成器（工具调用在生成器内部处理）
            return self._execute_stream_generator(messages, selected_llm=selected_llm, **api_params)
        
        # 非流式模式
        # 提取工具参数（使用原始工具对象）
        tools = api_params.get('original_tools') or api_params.get('tools')
        
        # 创建 LLM API 参数（移除内部参数）
        llm_api_params = {k: v for k, v in api_params.items() if k != 'original_tools'}
        
        # 如果有选定的LLM，临时替换self.llm
        original_llm = self.llm
        if selected_llm is not None:
            self.llm = selected_llm
        
        try:
            response = await super().execute(messages, **llm_api_params)
        finally:
            # 恢复原始LLM
            self.llm = original_llm
        
        # 从 ChatCompletion 对象中提取字符串内容
        if hasattr(response, 'choices') and hasattr(response.choices[0], 'message'):
            message = response.choices[0].message
            if hasattr(message, 'content') and message.content:
                response_text = message.content
            else:
                response_text = ""
        else:
            # 如果不是 ChatCompletion 对象，假设是字符串
            response_text = str(response)
        
        # 处理工具调用（仅在非流式模式）
        if tools is not None:
            response_text = await self._handle_tool_calls(response, tools, messages, llm_api_params)

        if isinstance(response_format, type) and issubclass(response_format, BaseModel):
            if format == "xml" or format == "xml_simple":
                # 清理响应中的代码块标记
                response_text = response_text.strip()
                if response_text.startswith('```xml'):
                    response_text = response_text[6:]  # 移除开头的 ```xml
                if response_text.endswith('```'):
                    response_text = response_text[:-3]  # 移除结尾的 ```
                response_text = response_text.strip()

                # 确保响应是有效的 XML
                if not response_text.strip().startswith('<?xml'):
                    response_text = f'<?xml version="1.0" encoding="UTF-8"?>\n{response_text}'

                # 根据format选择解析方式
                if format == "xml_simple":
                    response_text = self._simple_xml_to_json(response_format, response_text)
                else:
                    response_text = self._xml_to_json(response_text)
            response_text = self.normalize_response(response_text)

        if original_response_format and isinstance(original_response_format, type) and issubclass(original_response_format, BaseModel):
            response_text = original_response_format.model_validate_json(response_text)

        if self.output_parser:
            response_text = self.output_parser(response_text)
        return response_text
    
    async def _execute_stream_generator(self, messages, selected_llm=None, **api_params):
        """流式执行生成器，支持工具调用"""
        # 处理可选的 llm 参数
        current_llm = selected_llm if selected_llm is not None else self.llm
        
        # 添加 input_parser 处理
        if self.input_parser:
            messages = self.input_parser(messages)

        # Convert string/single message to list
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]
        elif isinstance(messages, Message):
            messages = [messages.model_dump()]
        elif isinstance(messages, dict) and "role" in messages:
            messages = [messages]
        elif isinstance(messages, list) and all(isinstance(msg, Message) for msg in messages):
            messages = [msg.model_dump() for msg in messages]

        # 提取工具参数（使用原始工具对象）
        tools = api_params.get('original_tools') or api_params.get('tools')
        
        # 使用流式生成
        full_content = ""
        tool_calls = []
        
        # 创建 LLM API 参数（移除内部参数）
        llm_api_params = {k: v for k, v in api_params.items() if k != 'original_tools'}
        
        # 确保stop参数被传递
        if 'stop' in api_params:
            llm_api_params['stop'] = api_params['stop']
        
        async for chunk in current_llm.generate_stream(messages, **llm_api_params):
            # 处理 StreamChunk 对象
            if hasattr(chunk, 'content'):
                if chunk.chunk_type == "text":
                    content = chunk.content
                    full_content += content
                    yield chunk  # 直接传递文本 StreamChunk 对象
                elif chunk.chunk_type == "tool_call":
                    # 收集工具调用
                    tool_call = chunk.metadata.get('tool_call')
                    if tool_call:
                        tool_calls.append(tool_call)
                    yield chunk  # 传递工具调用 StreamChunk
                    
                    # 立即执行工具并 yield 工具响应
                    if tool_call:
                        async for response_chunk in self._execute_and_yield_tool_response(tool_call, tools):
                            yield response_chunk
            else:
                # 向后兼容字符串
                content = str(chunk)
                full_content += content
                yield content
        
        # 如果有工具调用，递归获取最终响应
        if tools and tool_calls:
            # 构造包含工具调用的消息
            assistant_message = {
                "role": "assistant",
                "content": full_content if full_content else None,
                "tool_calls": tool_calls
            }
            
            # 执行所有工具调用并获取结果
            tool_results = []
            for tool_call in tool_calls:
                tool_result = await self._execute_single_tool_call(tool_call, tools)
                tool_results.append(tool_result)

            # 检查是否调用了 final_answer，如果是则不再递归调用 LLM
            has_final_answer = any(
                tc.get("function", {}).get("name") == "final_answer"
                for tc in tool_calls
            )

            if has_final_answer:
                # final_answer 已调用，直接结束，不再递归
                return

            # 添加工具调用消息和结果到对话历史
            updated_messages = messages + [assistant_message] + tool_results

            # 递归调用获取最终响应
            async for chunk in self._execute_stream_generator(updated_messages, **api_params):
                yield chunk
        else:
            pass
            # 最终处理
            # if self.output_parser:
            #     full_content = self.output_parser(full_content)
            #
            # # 返回最终完整内容作为特殊标记
            # yield f"[STREAM_COMPLETE: {full_content}]"
    
    async def _execute_and_yield_tool_response(self, tool_call, tools):
        """执行工具调用并立即 yield 工具响应"""
        tool_call_id = tool_call.get('id', 'unknown')
        function_name = tool_call.get('function', {}).get('name')
        function_args = tool_call.get('function', {}).get('arguments', '{}')
        
        # 查找对应的工具
        tool = self._find_tool(function_name, tools)
        
        if tool:
            try:
                # 解析参数
                import json
                args = json.loads(function_args) if function_args else {}
                
                # 执行工具 - 统一的调用方式
                if hasattr(tool, 'execute'):
                    # 传统的 execute 方法
                    result = tool.execute(**args)
                    import asyncio
                    if asyncio.iscoroutine(result):
                        result = await result
                elif callable(tool):
                    # 直接调用工具
                    result = tool(**args)
                    import asyncio
                    if asyncio.iscoroutine(result):
                        result = await result
                else:
                    result = f"Error: Tool {function_name} is not callable"
                
                # 立即 yield 工具响应
                from minion.main.action_step import StreamChunk

                # Check if this is the final_answer tool
                is_final = function_name == "final_answer"

                tool_response_chunk = StreamChunk(
                    content=f"\n📊 工具执行结果: {str(result)}\n",
                    chunk_type="tool_response" if not is_final else "final_answer",
                    metadata={
                        "tool_call_id": tool_call_id,
                        "tool_name": function_name,
                        "tool_result": str(result)
                    }
                )
                # Set is_final_answer flag for final_answer tool
                if is_final:
                    tool_response_chunk.is_final_answer = True

                yield tool_response_chunk
                
            except Exception as e:
                # 工具执行失败，也要 yield 错误响应
                error_msg = f"Error executing {function_name}: {str(e)}"
                
                from minion.main.action_step import StreamChunk
                error_response_chunk = StreamChunk(
                    content=f"\n❌ 工具执行错误: {error_msg}\n",
                    chunk_type="tool_response",
                    metadata={
                        "tool_call_id": tool_call_id,
                        "tool_name": function_name,
                        "error": str(e)
                    }
                )
                yield error_response_chunk
        else:
            # 工具未找到，也要 yield 错误响应
            error_msg = f"Error: Tool {function_name} not found"
            
            from minion.main.action_step import StreamChunk
            error_response_chunk = StreamChunk(
                content=f"\n❌ 工具未找到: {function_name}\n",
                chunk_type="tool_response",
                metadata={
                    "tool_call_id": tool_call_id,
                    "tool_name": function_name,
                    "error": error_msg
                }
            )
            yield error_response_chunk
    
    def _find_tool(self, function_name, tools):
        """查找指定名称的工具"""
        for tool in tools:
            if hasattr(tool, 'name') and tool.name == function_name:
                return tool
            elif hasattr(tool, 'get_schema'):
                schema = tool.get_schema()
                if schema.get('function', {}).get('name') == function_name:
                    return tool
        return None
    
    async def _execute_single_tool_call(self, tool_call, tools):
        """执行单个工具调用并返回结果消息"""
        tool_call_id = tool_call.get('id', 'unknown')
        function_name = tool_call.get('function', {}).get('name')
        function_args = tool_call.get('function', {}).get('arguments', '{}')
        
        # 查找对应的工具
        tool = self._find_tool(function_name, tools)
        
        if tool:
            try:
                # 解析参数
                import json
                args = json.loads(function_args) if function_args else {}
                
                # 执行工具 - 统一的调用方式
                if hasattr(tool, 'execute'):
                    # 传统的 execute 方法
                    result = tool.execute(**args)
                    import asyncio
                    if asyncio.iscoroutine(result):
                        result = await result
                elif callable(tool):
                    # 直接调用工具
                    result = tool(**args)
                    import asyncio
                    if asyncio.iscoroutine(result):
                        result = await result
                else:
                    result = f"Error: Tool {function_name} is not callable"
                
                # 创建工具结果消息
                return {
                    "role": "tool",
                    "tool_call_id": tool_call_id,
                    "content": str(result)
                }
                
            except Exception as e:
                # 工具执行失败
                error_msg = f"Error executing {function_name}: {str(e)}"
                return {
                    "role": "tool",
                    "tool_call_id": tool_call_id,
                    "content": error_msg
                }
        else:
            # 工具未找到
            error_msg = f"Error: Tool {function_name} not found"
            return {
                "role": "tool",
                "tool_call_id": tool_call_id,
                "content": error_msg
            }
    
    async def _handle_tool_calls_stream(self, response, tools, messages, api_params):
        """处理流式模式下的工具调用"""
        # 简化版本，直接返回响应
        # 在流式模式下，工具调用会在完整响应后处理
        return response

    def _dict_to_xml_example(self, data, root_name="root"):
        """Helper method to convert a dictionary to XML example string."""
        if isinstance(data, dict):
            elements = []
            for key, value in data.items():
                elements.append(f"<{key}>{self._dict_to_xml_example(value, key)}</{key}>")
            if root_name == "root":
                return f'<?xml version="1.0" encoding="UTF-8"?>\n<{root_name}>\n  {"  ".join(elements)}\n</{root_name}>'
            return "\n".join(elements)
        elif isinstance(data, list):
            elements = []
            item_name = root_name[:-1] if root_name.endswith('s') else 'item'
            for item in data:
                elements.append(f"<{item_name}>{self._dict_to_xml_example(item, item_name)}</{item_name}>")
            return "\n".join(elements)
        else:
            return str(data) if data is not None else ""

    def _xml_to_json(self, xml_str: str) -> str:
        """Convert XML string to JSON string compatible with Pydantic model."""
        # 移除 XML 声明
        xml_str = re.sub(r'<\?xml[^>]+\?>', '', xml_str).strip()

        # 解析 XML
        root = ET.fromstring(xml_str)

        # 如果根元素是 'root'，我们需要提取其子元素
        if root.tag == 'root':
            result = {}
            for child in root:
                result[child.tag] = self._process_xml_element(child)
        else:
            result = self._process_xml_element(root)

        return json.dumps(result)

    def _process_xml_element(self, element: ET.Element) -> Any:
        """递归处理 XML 元素"""
        # 如果元素没有子元素且有文本
        if len(element) == 0:
            text = element.text.strip() if element.text else ""
            # 尝试转换布尔值
            if text.lower() == 'true':
                return True
            elif text.lower() == 'false':
                return False
            # 尝试转换数字
            try:
                if '.' in text:
                    return float(text)
                return int(text)
            except ValueError:
                return text

        # 如果元素有子元素
        result = {}
        for child in element:
            # 处理列表情况（相同标签的多个元素）
            if child.tag in result:
                if not isinstance(result[child.tag], list):
                    result[child.tag] = [result[child.tag]]
                result[child.tag].append(self._process_xml_element(child))
            else:
                result[child.tag] = self._process_xml_element(child)

        return result

    def _simple_xml_to_json(self, response_format, xml_str: str) -> str:
        """使用简单的正则表达式解析XML"""
        # 移除XML声明
        xml_str = re.sub(r'<\?xml[^>]+\?>', '', xml_str).strip()

        # 获取schema中定义的所有字段
        schema = response_format.model_json_schema()
        fields = schema.get('properties', {}).keys()

        result = {}
        for field in fields:
            # 使用非贪婪匹配来提取标签内容
            pattern = f"<{field}>(.*?)</{field}>"
            match = re.search(pattern, xml_str, re.DOTALL)
            if match:
                value = match.group(1).strip()
                # 尝试转换布尔值
                if value.lower() == 'true':
                    result[field] = True
                elif value.lower() == 'false':
                    result[field] = False
                else:
                    # 尝试转换数字
                    try:
                        if '.' in value:
                            result[field] = float(value)
                        else:
                            result[field] = int(value)
                    except ValueError:
                        result[field] = value

        return json.dumps(result)
    
    def _format_tools_for_api(self, tools):
        """
        将工具列表转换为API调用格式
        
        Args:
            tools: 工具列表，可以是 BaseTool 实例列表
            
        Returns:
            list: 格式化后的工具定义列表，符合OpenAI API格式
        """
        if not tools:
            return []
            
        formatted_tools = []
        
        for tool in tools:
            # 检查是否是 BaseTool 实例
            if hasattr(tool, 'name') and hasattr(tool, 'description') and hasattr(tool, 'inputs'):
                tool_def = {
                    "type": "function",
                    "function": {
                        "name": tool.name,
                        "description": tool.description,
                        "parameters": {
                            "type": "object",
                            "properties": tool.inputs,
                            "required": [
                                name for name, param in tool.inputs.items() 
                                if not param.get("nullable", False)
                            ]
                        }
                    }
                }
                formatted_tools.append(tool_def)
            elif hasattr(tool, 'to_function_spec'):
                # 支持 MCP 工具的 to_function_spec 方法
                spec = tool.to_function_spec()
                if isinstance(spec, dict):
                    formatted_tools.append(spec)
            elif hasattr(tool, 'get_schema'):
                # 支持有 get_schema 方法的工具
                schema = tool.get_schema()
                if isinstance(schema, dict) and "function" in schema:
                    formatted_tools.append(schema)
                elif isinstance(schema, dict) and "type" in schema and schema["type"] == "function":
                    formatted_tools.append(schema)
            elif isinstance(tool, dict) and "function" in tool:
                # 如果已经是正确格式的工具定义
                formatted_tools.append(tool)
            elif callable(tool):
                decorated = decorate_tool(tool)
                tool_def = {
                    "type": "function",
                    "function": {
                        "name": decorated.name,
                        "description": decorated.description,
                        "parameters": {
                            "type": "object",
                            "properties": decorated.inputs,
                            "required": [
                                name for name, param in decorated.inputs.items()
                                if not param.get("nullable", False)
                            ]
                        }
                    }
                }
                formatted_tools.append(tool_def)
                     
        return formatted_tools
    
    async def _handle_tool_calls(self, response, tools, messages, api_params):
        """
        处理工具调用响应

        Args:
            response: LLM的原始响应（可能是字符串或 ChatCompletion 对象）
            tools: 可用的工具列表
            messages: 消息历史
            api_params: API参数

        Returns:
            str: 处理后的响应
        """
        # 检查是否是 ChatCompletion 对象
        if hasattr(response, 'choices') and hasattr(response.choices[0], 'message'):
            message = response.choices[0].message
            # 检查是否有 tool_calls
            if hasattr(message, 'tool_calls') and message.tool_calls:
                # 处理 OpenAI 格式的 tool_calls
                final_response = ""
                for tool_call in message.tool_calls:
                    tool_name = tool_call.function.name
                    args_str = tool_call.function.arguments

                    # 如果是final_answer，则直接抛出FinalAnswerException
                    if tool_name == 'final_answer':
                        try:
                            import json
                            args_dict = json.loads(args_str) if isinstance(args_str, str) else args_str
                            answer_value = args_dict.get('answer', str(args_dict))
                            raise FinalAnswerException(answer_value)
                        except json.JSONDecodeError:
                            # 如果解析失败，直接使用整个参数字符串
                            raise FinalAnswerException(args_str)

                    # 查找对应的工具
                    target_tool = None
                    for tool in tools:
                        if hasattr(tool, 'name') and tool.name == tool_name:
                            target_tool = tool
                            break
                        elif hasattr(tool, '__name__') and tool.__name__ == tool_name:
                            target_tool = tool
                            break

                    if target_tool:
                        try:
                            if callable(target_tool):
                                # 普通可调用对象
                                if isinstance(args_str,str):
                                    import json
                                    try:
                                        args_dict = json.loads(args_str)
                                        tool_result = target_tool(**args_dict)
                                    except:
                                        args_str.strip().strip('"\'')
                                        args_dict = json.loads(args_str)
                                        tool_result = target_tool(**args_dict)
                                else:
                                    tool_result = target_tool(**args_str) #assume it's dict

                                # 检查是否是 awaitable，如果是则 await
                                if inspect.iscoroutine(tool_result):
                                    tool_result = await tool_result
                            else:
                                tool_result = "Tool call failed: tool is not callable"

                            final_response += _format_tool_result(tool_name, tool_result) + "\n"

                        except FinalAnswerException as e:
                            # 特殊处理 FinalAnswerException
                            final_response += f"FINAL_ANSWER:{e.answer}\n"
                        except Exception as e:
                            error_msg = f"Tool {tool_name} execution error: {str(e)}"
                            final_response += error_msg + "\n"

                return final_response.strip()
            else:
                # 没有 tool_calls，返回正常内容
                return message.content or ""

        # 如果不是 ChatCompletion 对象，按原来的字符串处理方式
        import json
        import re

        # 提取响应文本内容用于正则匹配
        response_text = str(response) if not isinstance(response, str) else response

        # 首先尝试解析 XML 格式的工具调用
        xml_tool_call_pattern = r'<tool_call>\s*<tool_name>(\w+)</tool_name>\s*<parameters>(.*?)</parameters>\s*</tool_call>'
        xml_matches = re.finditer(xml_tool_call_pattern, response_text, re.IGNORECASE | re.DOTALL)

        final_response = response_text
        
        # 处理 XML 格式的工具调用
        for match in xml_matches:
            tool_name = match.group(1)
            parameters_xml = match.group(2).strip()
            
            # 查找对应的工具
            target_tool = None
            for tool in tools:
                if hasattr(tool, 'name') and tool.name == tool_name:
                    target_tool = tool
                    break
                elif hasattr(tool, '__name__') and tool.__name__ == tool_name:
                    target_tool = tool
                    break
            
            if target_tool:
                try:
                    # 解析参数
                    if hasattr(target_tool, 'forward'):
                        # BaseTool 实例，从XML中提取参数
                        if parameters_xml.strip():
                            # 从XML中提取参数值
                            # 如果是final_answer，则直接抛出FinalAnswerException，不调用forward
                            if tool_name == 'final_answer':
                                # 特殊处理 final_answer 工具
                                answer_match = re.search(r'<answer>(.*?)</answer>', parameters_xml, re.DOTALL)
                                if answer_match:
                                    answer_value = answer_match.group(1).strip()
                                    # 直接抛出异常，不调用 forward
                                    raise FinalAnswerException(answer_value)
                                else:
                                    # 直接抛出异常，不调用 forward
                                    raise FinalAnswerException(parameters_xml)
                            else:
                                # 其他工具的参数解析
                                tool_result = target_tool.forward(parameters_xml)
                        else:
                            tool_result = target_tool.forward()
                        
                        # 检查是否是 awaitable，如果是则 await
                        if inspect.iscoroutine(tool_result):
                            tool_result = await tool_result
                    else:
                        tool_result = "Tool call failed: tool is not callable"

                    # 替换响应中的工具调用为工具结果
                    final_response = final_response.replace(
                        match.group(0),
                        _format_tool_result(tool_name, tool_result)
                    )

                except FinalAnswerException as e:
                    # 特殊处理 FinalAnswerException
                    final_response = final_response.replace(
                        match.group(0),
                        f"FINAL_ANSWER:{e.answer}"
                    )
                except Exception as e:
                    error_msg = f"Tool {tool_name} execution error: {str(e)}"
                    final_response = final_response.replace(match.group(0), error_msg)

        # 然后尝试解析传统的函数调用模式
        tool_call_pattern = r'(?:调用工具|使用工具|call tool|use tool).*?(\w+)\s*\((.*?)\)'
        matches = re.finditer(tool_call_pattern, final_response, re.IGNORECASE | re.DOTALL)
        
        for match in matches:
            tool_name = match.group(1)
            args_str = match.group(2)
            
            # 查找对应的工具
            target_tool = None
            for tool in tools:
                if hasattr(tool, 'name') and tool.name == tool_name:
                    target_tool = tool
                    break
                elif hasattr(tool, '__name__') and tool.__name__ == tool_name:
                    target_tool = tool
                    break
            
            if target_tool:
                try:
                    # 解析参数
                    if hasattr(target_tool, 'forward'):
                        # BaseTool 实例
                        if args_str.strip():
                            # 尝试解析为字符串参数
                            args_str = args_str.strip().strip('"\'')
                            tool_result = target_tool.forward(args_str)
                        else:
                            tool_result = target_tool.forward()
                        
                        # 检查是否是 awaitable，如果是则 await
                        if inspect.iscoroutine(tool_result):
                            tool_result = await tool_result
                            
                    elif callable(target_tool):
                        # 普通可调用对象
                        if args_str.strip():
                            args_str = args_str.strip().strip('"\'')
                            tool_result = target_tool(args_str)
                        else:
                            tool_result = target_tool()
                        
                        # 检查是否是 awaitable，如果是则 await
                        if inspect.iscoroutine(tool_result):
                            tool_result = await tool_result
                    else:
                        tool_result = "Tool call failed: tool is not callable"

                    # 替换响应中的工具调用为工具结果
                    final_response = final_response.replace(
                        match.group(0),
                        _format_tool_result(tool_name, tool_result)
                    )

                except FinalAnswerException as e:
                    # 特殊处理 FinalAnswerException
                    final_response = final_response.replace(
                        match.group(0),
                        f"FINAL_ANSWER:{e.answer}"
                    )
                except Exception as e:
                    error_msg = f"Tool {tool_name} execution error: {str(e)}"
                    final_response = final_response.replace(match.group(0), error_msg)
        
        return final_response



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Bash command execution tool
"""

import os
import subprocess
from typing import Optional, Any
from minion.tools import BaseTool
from minion.utils.output_truncator import truncate_output


class BashTool(BaseTool):
    """Bash command execution tool"""

    name = "bash"
    description = "Execute bash commands"
    readonly = False  # Command execution may modify system state
    inputs = {
        "command": {"type": "string", "description": "Bash command to execute"},
        "timeout": {
            "type": "integer",
            "description": "Timeout in seconds",
            "nullable": True,
        },
    }
    output_type = "string"

    def forward(self, command: str, timeout: Optional[int] = 30) -> str:
        """Execute bash command"""
        try:
            # Security check: prohibit dangerous commands
            dangerous_commands = ["rm -rf", "sudo", "su", "chmod 777", "mkfs", "dd if="]
            if any(dangerous in command.lower() for dangerous in dangerous_commands):
                return f"Error: Dangerous command prohibited - {command}"

            result = subprocess.run(
                command,
                shell=True,
                capture_output=True,
                text=True,
                timeout=timeout,
                cwd=os.getcwd(),
            )

            output = ""
            if result.stdout:
                output += f"Standard output:\n{result.stdout}\n"
            if result.stderr:
                output += f"Standard error:\n{result.stderr}\n"
            output += f"Exit code: {result.returncode}"

            return self.format_for_observation(output)

        except subprocess.TimeoutExpired:
            return f"Command execution timeout ({timeout} seconds)"
        except Exception as e:
            return f"Error executing command: {str(e)}"

    def format_for_observation(self, output: Any) -> str:
        """Format output, automatically truncate large content"""
        if isinstance(output, str):
            return truncate_output(output, tool_name=self.name)
        return str(output)



from typing import List, Optional, Dict, Any
from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.models import SystemMessage, UserMessage, AssistantMessage
from azure.core.credentials import AzureKeyCredential

from minion.logs import log_llm_stream
from minion.providers.base_provider import BaseProvider
from minion.providers.llm_provider_registry import llm_registry
from minion.schema.message_types import Message, MessageContent, ContentType
from minion.providers.openai_provider import OpenAIProvider


@llm_registry.register("azure_inference")
class AzureInferenceProvider(OpenAIProvider):
    def _setup(self) -> None:
        """Setup Azure Inference SDK client"""
        endpoint = self.config.base_url
        key = self.config.api_key
        self.model = self.config.model or self.config.deployment_name
        
        self.client = ChatCompletionsClient(
            endpoint=endpoint,
            credential=AzureKeyCredential(key)
        )
        # Also use the same client for sync operations since Azure Inference SDK
        # doesn't have separate sync/async clients
        self.client_sync = self.client

    # def _prepare_messages(self, messages: List[Message] | Message | str) -> List[Any]:
    #     """Convert minion Message objects to Azure Inference SDK message objects"""
    #     # Convert single message or string to list
    #     if isinstance(messages, (str, Message)):
    #         messages = [messages if isinstance(messages, Message) else Message(role="user", content=messages)]
    #
    #     azure_messages = []
    #     for msg in messages:
    #         content = msg.content
    #         if isinstance(content, str):
    #             text_content = content
    #         elif isinstance(content, MessageContent):
    #             if content.type == ContentType.TEXT:
    #                 text_content = content.text
    #             else:
    #                 # For now, we only handle text content for Azure Inference
    #                 # TODO: Add support for image content if Azure Inference SDK supports it
    #                 text_content = content.text if content.text else ""
    #         else:
    #             text_content = str(content)
    #
    #         if msg.role == "system":
    #             azure_messages.append(SystemMessage(content=text_content))
    #         elif msg.role == "user":
    #             azure_messages.append(UserMessage(content=text_content))
    #         elif msg.role == "assistant":
    #             azure_messages.append(AssistantMessage(content=text_content))
    #     return azure_messages

    async def generate(self, messages: List[Message], temperature: Optional[float] = None, **kwargs) -> str:
        """Generate completion using Azure Inference SDK"""
        azure_messages = self._prepare_messages(messages)
        
        # Prepare parameters
        params = {
            "messages": azure_messages,
            "model": self.model,
            "temperature": temperature if temperature is not None else 0.6,
            "max_tokens": kwargs.get("max_tokens", 1000)
        }
        
        # Add optional parameters if provided
        if "top_p" in kwargs:
            params["top_p"] = kwargs["top_p"]
        if "frequency_penalty" in kwargs:
            params["frequency_penalty"] = kwargs["frequency_penalty"]
        if "presence_penalty" in kwargs:
            params["presence_penalty"] = kwargs["presence_penalty"]
        
        response = self.client.complete(**params)
        return response.messages[0].content

    async def generate_stream(self, messages: List[Message], temperature: Optional[float] = None, **kwargs) -> str:
        """Generate streaming completion using Azure Inference SDK"""
        azure_messages = self._prepare_messages(messages)
        
        # Prepare parameters
        params = {
            "messages": azure_messages,
            "model": self.model,
            "temperature": temperature if temperature is not None else 0.6,
            "max_tokens": kwargs.get("max_tokens", 1000),
            "stream": True
        }
        
        # Add optional parameters if provided
        if "top_p" in kwargs:
            params["top_p"] = kwargs["top_p"]
        if "frequency_penalty" in kwargs:
            params["frequency_penalty"] = kwargs["frequency_penalty"]
        if "presence_penalty" in kwargs:
            params["presence_penalty"] = kwargs["presence_penalty"]
        
        response = self.client.complete(**params)
        full_content = ""
        completion_tokens = 0
        for chunk in response:
            if chunk.choices:
                completion_tokens += 1
                chunk_message = chunk.choices[0].delta.content
                if chunk_message:
                    full_content += chunk_message
                    # log_llm_stream(chunk_message)
                #yield chunk_message
        return full_content
        
    def generate_sync(self, messages: List[Message], temperature: Optional[float] = None, **kwargs) -> str:
        """Generate completion synchronously using Azure Inference SDK"""
        azure_messages = self._prepare_messages(messages)
        
        # Prepare parameters
        params = {
            "messages": azure_messages,
            "model": self.model,
            "temperature": temperature if temperature is not None else 0.6,
            "max_tokens": kwargs.get("max_tokens", 1000)
        }
        
        # Add optional parameters if provided
        if "top_p" in kwargs:
            params["top_p"] = kwargs["top_p"]
        if "frequency_penalty" in kwargs:
            params["frequency_penalty"] = kwargs["frequency_penalty"]
        if "presence_penalty" in kwargs:
            params["presence_penalty"] = kwargs["presence_penalty"]
        
        response = self.client_sync.complete(**params)
        return response.messages[0].content


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@Time    : 2024
@Author  : femto Zheng
@File    : tool_decorator.py
"""

import asyncio
import inspect
import json
import textwrap
import warnings
import ast
from functools import wraps
from typing import Callable, Union, Dict, Any, get_type_hints, Set

from .base_tool import BaseTool
from .async_base_tool import AsyncBaseTool


def get_json_schema(func: Callable) -> Dict[str, Any]:
    """从函数提取JSON schema定义"""
    sig = inspect.signature(func)
    doc = inspect.getdoc(func) or ""

    # 获取函数名和描述
    description = doc.split('\n\n')[0] if doc else ""

    # 从docstring解析参数描述
    param_descriptions = {}
    if 'Args:' in doc:
        args_section = doc.split('Args:')[1].split('Returns:')[0] if 'Returns:' in doc else doc.split('Args:')[1]
        for line in args_section.strip().split('\n'):
            if ':' in line:
                param_name = line.split(':')[0].strip()
                param_desc = line.split(':', 1)[1].strip()
                param_descriptions[param_name] = param_desc

    # 获取类型提示
    type_hints = get_type_hints(func)

    # 构建参数属性
    properties = {}
    for name, param in sig.parameters.items():
        if name in type_hints:
            param_type = type_hints[name]
            type_name = str(param_type)

            # 转换类型名到JSON schema类型
            json_type = "string"
            schema_def = {"type": json_type}

            if "str" in type_name:
                json_type = "string"
            elif "int" in type_name:
                json_type = "integer"
            elif "float" in type_name:
                json_type = "number"
            elif "bool" in type_name:
                json_type = "boolean"
            elif "list" in type_name or "List" in type_name:
                json_type = "array"
                schema_def = {
                    "type": "array",
                    "items": {"type": "string"}  # 默认数组项类型
                }
            elif "dict" in type_name or "Dict" in type_name:
                json_type = "object"
                schema_def = {
                    "type": "object",
                    "additionalProperties": True
                }

            # 检查参数是否可选
            is_optional = "Optional" in type_name or param.default != inspect.Parameter.empty

            # 使用改进的schema定义
            if isinstance(schema_def, dict) and "type" in schema_def:
                properties[name] = schema_def.copy()
                properties[name]["description"] = param_descriptions.get(name, "")
            else:
                properties[name] = {
                    "type": json_type,
                    "description": param_descriptions.get(name, ""),
                }

            if is_optional:
                properties[name]["nullable"] = True

    # 获取返回类型
    return_type = "any"
    if "return" in type_hints:
        return_type_str = str(type_hints["return"])
        if "str" in return_type_str:
            return_type = "string"
        elif "int" in return_type_str:
            return_type = "integer"
        elif "float" in return_type_str:
            return_type = "number"
        elif "bool" in return_type_str:
            return_type = "boolean"
        elif "list" in return_type_str or "List" in return_type_str:
            return_type = "array"
        elif "dict" in return_type_str or "Dict" in return_type_str:
            return_type = "object"

    return {
        "function": {
            "name": func.__name__,
            "description": description,
            "parameters": {
                "type": "object",
                "properties": properties,
                "required": [
                    name for name, param in sig.parameters.items()
                    if param.default == inspect.Parameter.empty
                ]
            },
            "return": {
                "type": return_type
            }
        }
    }


def get_imports(code: str) -> Set[str]:
    """获取代码中的导入模块"""
    imports = set()
    try:
        tree = ast.parse(code)
        for node in ast.walk(tree):
            if isinstance(node, ast.Import):
                for name in node.names:
                    imports.add(name.name.split('.')[0])
            elif isinstance(node, ast.ImportFrom):
                if node.module:
                    imports.add(node.module.split('.')[0])
    except SyntaxError:
        # 语法错误时返回空集合
        pass
    return imports


def readonly(func: Callable) -> Callable:
    """
    装饰器，将函数标记为只读工具
    
    Args:
        func: 要标记为只读的函数
        
    Returns:
        Callable: 标记了_readonly=True的函数
    """
    func._readonly = True
    return func


class TypeHintParsingException(Exception):
    """工具类型提示解析异常"""
    pass


def tool(tool_function: Callable) -> Union[BaseTool, AsyncBaseTool]:
    """
    统一装饰器，将函数转换为相应的工具实例
    - 同步函数转换为BaseTool子类
    - 异步函数转换为AsyncBaseTool子类
    
    Args:
        tool_function (Callable): 要转换为工具的函数
            应该为每个输入和输出提供类型提示
            应该有包含函数描述和'Args:'部分的文档字符串
            
    Returns:
        Union[BaseTool, AsyncBaseTool]: 基于函数创建的工具实例
    """
    # 检查是否为异步函数
    is_async = asyncio.iscoroutinefunction(tool_function)
    
    # 获取JSON schema
    tool_json_schema = get_json_schema(tool_function)["function"]
    
    # 检查函数是否需要state参数
    sig = inspect.signature(tool_function)
    type_hints = get_type_hints(tool_function)
    needs_state = False
    
    # 检查是否有state参数且类型为AgentState
    for param_name, param in sig.parameters.items():
        if param_name == 'state' and param_name in type_hints:
            type_hint = type_hints[param_name]
            # 检查类型提示是否包含AgentState
            type_str = str(type_hint)
            if 'AgentState' in type_str:
                needs_state = True
                break
    
    # 检查返回类型
    if "return" not in tool_json_schema:
        if len(tool_json_schema["parameters"]["properties"]) == 0:
            tool_json_schema["return"] = {"type": "null"}
        else:
            raise TypeHintParsingException(
                "Tool return type not found: make sure your function has a return type hint!"
            )
    
    if is_async:
        # 创建异步工具类
        @wraps(tool_function)
        async def wrapped_async_function(self, *args, **kwargs):
            return await tool_function(*args, **kwargs)
        
        class SimpleAsyncTool(AsyncBaseTool):
            def __init__(self):
                super().__init__()
                self.is_initialized = True
            
            # 直接定义forward方法
            forward = wrapped_async_function
        
        # 设置类属性
        SimpleAsyncTool.name = tool_json_schema["name"]
        SimpleAsyncTool.description = tool_json_schema["description"]
        SimpleAsyncTool.inputs = tool_json_schema["parameters"]["properties"]
        SimpleAsyncTool.output_type = tool_json_schema["return"]["type"]
        readonly_value = getattr(tool_function, '_readonly', False)
        SimpleAsyncTool.readonly = readonly_value
        SimpleAsyncTool.needs_state = needs_state
        # Debug print
        # print(f"DEBUG: Setting readonly={readonly_value} for async {tool_json_schema['name']}")
        
        # 设置输出schema（如果存在）
        if "output_schema" in tool_json_schema:
            SimpleAsyncTool.output_schema = tool_json_schema["output_schema"]
        elif "return" in tool_json_schema and "schema" in tool_json_schema["return"]:
            SimpleAsyncTool.output_schema = tool_json_schema["return"]["schema"]
        
        # 获取签名参数并添加self
        sig = inspect.signature(tool_function)
        new_sig = sig.replace(
            parameters=[inspect.Parameter("self", inspect.Parameter.POSITIONAL_OR_KEYWORD)] + list(sig.parameters.values())
        )
        SimpleAsyncTool.forward.__signature__ = new_sig
        
        # 创建并附加源代码
        _attach_source_code(SimpleAsyncTool, tool_function, tool_json_schema, is_async=True)
        
        return SimpleAsyncTool()
    
    else:
        # 创建同步工具类
        @wraps(tool_function)
        def wrapped_function(self, *args, **kwargs):
            return tool_function(*args, **kwargs)
        
        class SimpleTool(BaseTool):
            def __init__(self):
                super().__init__()
                self.is_initialized = True
            
            # 直接定义forward方法
            forward = wrapped_function
        
        # 设置类属性
        SimpleTool.name = tool_json_schema["name"]
        SimpleTool.description = tool_json_schema["description"]
        SimpleTool.inputs = tool_json_schema["parameters"]["properties"]
        SimpleTool.output_type = tool_json_schema["return"]["type"]
        readonly_value = getattr(tool_function, '_readonly', False)
        SimpleTool.readonly = readonly_value
        SimpleTool.needs_state = needs_state
        # Debug print
        # print(f"DEBUG: Setting readonly={readonly_value} for {tool_json_schema['name']}")
        
        # 设置输出schema（如果存在）
        if "output_schema" in tool_json_schema:
            SimpleTool.output_schema = tool_json_schema["output_schema"]
        elif "return" in tool_json_schema and "schema" in tool_json_schema["return"]:
            SimpleTool.output_schema = tool_json_schema["return"]["schema"]
        
        # 获取签名参数并添加self
        sig = inspect.signature(tool_function)
        new_sig = sig.replace(
            parameters=[inspect.Parameter("self", inspect.Parameter.POSITIONAL_OR_KEYWORD)] + list(sig.parameters.values())
        )
        SimpleTool.forward.__signature__ = new_sig
        
        # 创建并附加源代码
        _attach_source_code(SimpleTool, tool_function, tool_json_schema, is_async=False)
        
        return SimpleTool()


def _attach_source_code(tool_class, tool_function, tool_json_schema, is_async=False):
    """
    为动态创建的工具类和forward方法附加源代码
    
    Args:
        tool_class: 动态创建的工具类
        tool_function: 原始工具函数
        tool_json_schema: 工具的JSON schema
        is_async: 是否为异步函数
    """
    try:
        # 获取工具函数的源代码
        tool_source = textwrap.dedent(inspect.getsource(tool_function))
        
        # 移除装饰器和函数定义行
        lines = tool_source.splitlines()
        tree = ast.parse(tool_source)
        
        # 查找函数定义（包括异步函数）
        func_node = next((node for node in ast.walk(tree) 
                         if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef))), None)
        if not func_node:
            raise ValueError(
                f"No function definition found in the provided source of {tool_function.__name__}. "
                "Ensure the input is a standard function."
            )
        
        # 提取装饰器行
        decorator_lines = ""
        if func_node.decorator_list:
            tool_decorators = [d for d in func_node.decorator_list if isinstance(d, ast.Name) and d.id == "tool"]
            if len(tool_decorators) > 1:
                raise ValueError(
                    f"Multiple @tool decorators found on function '{func_node.name}'. Only one @tool decorator is allowed."
                )
            if len(tool_decorators) < len(func_node.decorator_list):
                warnings.warn(
                    f"Function '{func_node.name}' has decorators other than @tool. "
                    "This may cause issues with serialization in the remote executor. See issue #1626."
                )
            if tool_decorators:
                decorator_start = tool_decorators[0].lineno - 1
                decorator_end = func_node.decorator_list[-1].lineno - 1
                decorator_lines = "\n".join(lines[decorator_start:decorator_end + 1])
        
        # 提取工具源代码主体
        body_start = func_node.body[0].lineno - 1  # AST lineno从1开始
        tool_source_body = "\n".join(lines[body_start:])
        
        # 获取新的签名
        new_sig = tool_class.forward.__signature__
        
        # 创建forward方法源代码，包括def行和缩进
        if is_async:
            forward_method_source = f"async def forward{new_sig}:\n{tool_source_body}"
            base_class_name = "AsyncBaseTool"
        else:
            forward_method_source = f"def forward{new_sig}:\n{tool_source_body}"
            base_class_name = "BaseTool"
        
        # 创建类源代码
        indent = " " * 4  # 类方法缩进
        class_source = (
            textwrap.dedent(f'''
            class SimpleTool({base_class_name}):
                name: str = "{tool_json_schema["name"]}"
                description: str = {json.dumps(textwrap.dedent(tool_json_schema["description"]).strip())}
                inputs: dict[str, dict[str, str]] = {tool_json_schema["parameters"]["properties"]}
                output_type: str = "{tool_json_schema["return"]["type"]}"

                def __init__(self):
                    self.is_initialized = True

            ''')
            + textwrap.indent(decorator_lines, indent)
            + textwrap.indent(forward_method_source, indent)
        )
        
        # 存储源代码到类和方法上以供检查
        tool_class.__source__ = class_source
        tool_class.forward.__source__ = forward_method_source
        
    except (IOError, TypeError, OSError):
        # 无法获取源代码时设置为None
        tool_class.__source__ = None
        tool_class.forward.__source__ = None


from typing import List, Optional, AsyncIterator

from minion.providers.openai_provider import OpenAIProvider
from minion.providers.llm_provider_registry import llm_registry
from minion.schema.message_types import Message


@llm_registry.register("azure")
class AzureProvider(OpenAIProvider):
    def _setup(self) -> None:
        import openai
        
        # Azure OpenAI 需要特定的配置
        client_kwargs = {
            "api_key": self.config.api_key,
            "azure_endpoint": self.config.base_url,  # Azure 使用 endpoint 而不是 base_url
            "api_version": self.config.api_version or "2024-05-01-preview",  # Azure 需要 api_version
        }
        
        # Azure OpenAI 使用 azure_deployment_name 而不是 model
        if hasattr(self.config, "deployment_name"):
            client_kwargs["azure_deployment"] = self.config.deployment_name
        else:
            client_kwargs["azure_deployment"] = self.config.model
            
        self.client_sync = openai.AzureOpenAI(**client_kwargs)
        self.client = openai.AsyncAzureOpenAI(**client_kwargs)

    async def generate(self, messages: List[Message], temperature: Optional[float] = None, **kwargs) -> str:
        # 移除 model 参数，因为 Azure 使用 deployment_name
        #kwargs.pop("model", None)
        return await super().generate(messages, temperature, **kwargs)

    async def generate_stream(self, messages: List[Message], temperature: Optional[float] = None, **kwargs):
        # 移除 model 参数，因为 Azure 使用 deployment_name
        #kwargs.pop("model", None)
        async for chunk in super().generate_stream(messages, temperature, **kwargs):
            yield chunk
        
    def generate_sync(self, messages: List[Message], temperature: Optional[float] = None, **kwargs) -> str:
        # 移除 model 参数，因为 Azure 使用 deployment_name
        #kwargs.pop("model", None)
        return super().generate_sync(messages, temperature, **kwargs) 